/*
** dump.c - utility to display or modify contents of memory windows on an
**          Alpha Data Gen 3 reconfigurable computing device.
**
** (C) Copyright 2009-2014 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>
# include <limits.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _stprintf_s snprintf
#   define _stscanf_s(buffer, ...) _stscanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
#   define _tcscpy_s(x, y, z) tcscpy(x, z)
#   define _tcsnset_s(buffer, max, c, n) _tcsnset(buffer, c, n)
#endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _istprint(x) isprint(x)
# define _stprintf_s snprintf
# define _stscanf_s sscanf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# if defined(__VXWORKS__) || defined(__vxworks)
#   define _tcsicmp(pStr1, pStr2) ourStrCaseCmp(pStr1, pStr2)
# else
#   define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# endif
# define _tcslen(x) strlen(x)
#   define _tcsnset_s(buffer, max, c, n) strset_s(buffer, max, c, n)
# define _tmain main
# define _tprintf printf
# define _tscanf(buffer, ...) scanf(buffer, __VA_ARGS__)
# define _totlower(c) tolower(c)
# define _gettc getc
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

#if defined(_UNICODE)
# define EOF_VALUE WEOF
#else
# define EOF_VALUE EOF
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define MIN(a, b) ((a) < (b) ? (a) : (b))

#define MAX_TOKEN_LENGTH (128)

#define FILL_BUFFER_SIZE (0x10000U)

typedef enum _CommandType {
  CommandTypeFill,
  CommandTypeRead,
  CommandTypeWrite
} CommandType;

typedef enum _TokenSubstate {
  NotInToken,
  InTokenDecimal,
  InTokenHexadecimal,
  InTokenInvalid
} TokenSubstate;

typedef struct _TokenState {
  unsigned int position;
  TokenSubstate substate;
  unsigned long long value;
  TCHAR tokenBuf[MAX_TOKEN_LENGTH];
} TokenState;

static const TokenState g_tokenStateInit = {
  0, NotInToken
};

#if !defined(_WIN32)
static void
strset_s(
  TCHAR* pBuffer,
  size_t max,
  TCHAR c,
  size_t n)
{
  if (!max) {
    return;
  }
  max--;
  while (max) {
    *pBuffer++ = c;
    max--;
  }
  *pBuffer++ = _T('\0');
}
#endif

#if defined(__VXWORKS__) || defined(__vxworks)
static int
ourStrCaseCmp(
  const char* pS1,
  const char* pS2)
{
  char c1, c2;

  while (1) {
    c1 = *pS1;
    c2 = *pS2;
    if (c1 == '\0' || tolower(c1) != tolower(c2)) {
      break;
    }
    pS1++;
    pS2++;
  }
  return tolower(c1) - tolower(c2);
}
#endif

static bool_t
fillMemoryUnaligned(
  uint8_t* p,
  unsigned int width,
  bool_t bBigEndian,
  uint64_t fillValue,
	size_t n)
{
  union {
    uint8_t asUint8[FILL_BUFFER_SIZE / sizeof(uint8_t)];
    uint16_t asUint16[FILL_BUFFER_SIZE / sizeof(uint16_t)];
    uint32_t asUint32[FILL_BUFFER_SIZE / sizeof(uint32_t)];
    uint64_t asUint64[FILL_BUFFER_SIZE / sizeof(uint64_t)];
  } fillBuffer;
  size_t i, m = n / width, chunk;
  uint8_t* p8;

  chunk = n > FILL_BUFFER_SIZE ? FILL_BUFFER_SIZE : n;
  m = chunk / width;
  switch (width) {
  case 1:
    if (bBigEndian) {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint8[i] = cpu_to_be8((uint8_t)fillValue);
      }
    } else {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint8[i] = cpu_to_le8((uint8_t)fillValue);
      }
    }
    break;

  case 2:
    if (bBigEndian) {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint16[i] = cpu_to_be16((uint16_t)fillValue);
      }
    } else {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint16[i] = cpu_to_le16((uint16_t)fillValue);
      }
    }
    break;

  case 4:
    if (bBigEndian) {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint32[i] = cpu_to_be32((uint32_t)fillValue);
      }
    } else {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint32[i] = cpu_to_le32((uint32_t)fillValue);
      }
    }
    break;

  case 8:
    if (bBigEndian) {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint64[i] = cpu_to_be64(fillValue);
      }
    } else {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint64[i] = cpu_to_le64(fillValue);
      }
    }
    break;

  default:
    assert(FALSE);
  }

  p8 = (uint8_t*)p;
  while (n) {
    chunk = n > FILL_BUFFER_SIZE ? FILL_BUFFER_SIZE : n;
    memcpy(p, fillBuffer.asUint8, chunk);
    n -= chunk;
    p8 += chunk;
  }

  return TRUE; /* OK */
}

static bool_t
fillMemory(
  uint8_t* p,
  unsigned int width,
  bool_t bBigEndian,
  uint64_t fillValue,
	size_t n)
{
  size_t i;

  if ((uintptr_t)p & (width - 1U)) {
    return fillMemoryUnaligned(p, width, bBigEndian, fillValue, n);
  }

  switch (width) {
  case 1:
    if (bBigEndian) {
      uint8_t* p8 = (uint8_t*)p;

      for (i = 0; i < n; i += width) {
        *p8++ = cpu_to_be8((uint8_t)fillValue);
      }
    } else {
      uint8_t* p8 = (uint8_t*)p;

      for (i = 0; i < n; i += width) {
        *p8++ = cpu_to_le8((uint8_t)fillValue);
      }
    }
    break;

  case 2:
    if (bBigEndian) {
      uint16_t* p16 = (uint16_t*)p;

      for (i = 0; i < n; i += width) {
        *p16++ = cpu_to_be16((uint16_t)fillValue);
      }
    } else {
      uint16_t* p16 = (uint16_t*)p;

      for (i = 0; i < n; i += width) {
        *p16++ = cpu_to_le16((uint16_t)fillValue);
      }
    }
    break;

  case 4:
    if (bBigEndian) {
      uint32_t* p32 = (uint32_t*)p;

      for (i = 0; i < n; i += width) {
        *p32++ = cpu_to_be32((uint32_t)fillValue);
      }
    } else {
      uint32_t* p32 = (uint32_t*)p;

      for (i = 0; i < n; i += width) {
        *p32++ = cpu_to_le32((uint32_t)fillValue);
      }
    }
    break;

  case 8:
    if (bBigEndian) {
      uint64_t* p64 = (uint64_t*)p;

      for (i = 0; i < n; i += width) {
        *p64++ = cpu_to_be64(fillValue);
      }
    } else {
      uint64_t* p64 = (uint64_t*)p;

      for (i = 0; i < n; i += width) {
        *p64++ = cpu_to_le64(fillValue);
      }
    }
    break;

  default:
    assert(FALSE);
  }

  return TRUE; /* OK */
}

static bool_t
dumpMemoryViaWindow(
	unsigned int indent,
	unsigned int nColumn,
	unsigned int radix,
  bool_t bBigEndian,
	bool_t bAlign,
  uint64_t offset,
	const void* pBuffer,
	size_t n)
{
  static const unsigned int offsetSize = sizeof(uint64_t);
  unsigned int offsetStrLen = 3 + 2 * offsetSize + (offsetSize - 1) / 4;
	uint64_t address, limit, addressCol, m;
	unsigned int addressLo, columnMask;
	uint8_t* p;
	uint8_t* p8;
	TCHAR line[257];
	TCHAR* pPos;
	uint8_t columnBuf[256];
	uint8_t val8;
	uint16_t val16;
	uint32_t val32;
	uint64_t val64;
	unsigned int i, j, k, chunk;
  unsigned int rowBytes;
  int nwritten;
  size_t remaining;

	if (indent > 128) {
		_tprintf(_T("*** dumpMemoryViaWindow: invalid value '%lu' for indent\n"), (unsigned long)indent);
		return FALSE; /* Failure */
	}
  /* SIZE_T_MAX should really be used here, but unfortunately it is not very portable */
  if (offset > (size_t)-1) {
		_tprintf(_T("*** dumpMemoryViaWindow: offset value '0x%llX' larger than SIZE_T_MAX\n"), (unsigned long long)offset);
		return FALSE; /* Failure */
  }
  switch (nColumn) {
  case 1:
  case 2:
  case 4:
  case 8:
  case 16:
  case 32:
    break;
  default:
    _tprintf(_T("*** dumpMemoryViaWindow: invalid value '%lu' for nColumn\n"), (unsigned long)nColumn);
    return FALSE; /* Failure */
  }
	if (radix != 1 && radix != 2 && radix != 4 && radix != 8) {
		_tprintf(_T("*** dumpMemoryViaWindow: invalid value '%lu' for radix\n"), (unsigned long)radix);
		return FALSE; /* Failure */
	}

  rowBytes = nColumn * radix;

  /* Make sure everything is aligned */
	columnMask = rowBytes - 1U;
	address = (uint64_t)(uintptr_t)pBuffer;
	limit = address + n;
	if (bAlign) {
		addressCol = offset & ~(uint64_t)columnMask;
		addressLo = ((unsigned int)address) & columnMask;
	} else {
		addressCol = offset;
		addressLo = 0;
	}

	m = limit - address;
	p = (uint8_t*)pBuffer;

	pPos = line;
	pPos[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent);
  pPos += MIN(indent, remaining);
  remaining -= MIN(indent, remaining);
  nwritten = _stprintf_s(pPos, remaining, _T("Dump of memory at %p + %lu(0x%lX) bytes:\n"),
		pBuffer, (unsigned long)n, (unsigned long)n);
  if (nwritten > 0) {
    pPos += nwritten;
    remaining -= nwritten;
  }
  line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
	_tprintf(_T("%s"), line);

	pPos = line;
  line[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent + 1 + offsetStrLen);
  pPos += MIN(indent + 1 + offsetStrLen, remaining);
  remaining -= MIN(indent + 1 + offsetStrLen, remaining);
	for (i = 0; i < rowBytes; i += radix) {
		switch (radix) {
    case 0x1U:
      nwritten = _stprintf_s(pPos, remaining, _T("%02lX "), (unsigned long)i);
      break;

    case 0x2U:
      nwritten = _stprintf_s(pPos, remaining, _T("  %02lX "), (unsigned long)i);
      break;

    case 0x4U:
      nwritten = _stprintf_s(pPos, remaining, _T("      %02lX "), (unsigned long)i);
      break;

    case 0x8U:
      nwritten = _stprintf_s(pPos, remaining, _T("               %02lX "), (unsigned long)i);
      break;
    }
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }
	}
  if (remaining) {
	  *pPos++ = _T('\n');
    remaining--;
  }
  if (remaining) {
	  *pPos++ = _T('\0');
    remaining--;
  }
  line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
	_tprintf(_T("%s"), line);

	while (m) {
		chunk = rowBytes - addressLo;
		if (chunk > m) {
			chunk = (unsigned int)m;
		}

		memset(columnBuf, 0, rowBytes);
		memcpy(columnBuf + addressLo, p, chunk);

		p8 = (uint8_t*)columnBuf;

		pPos = line;
    line[0] = _T('\0');
    remaining = ARRAY_LENGTH(line);
    _tcsnset_s(pPos, remaining, _T(' '), indent);
		pPos += MIN(indent, remaining);
    remaining -= MIN(indent, remaining);
		nwritten = _stprintf_s(pPos, remaining, _T("0x%08lX_%08lX: "), (unsigned long)(uint32_t)(addressCol >> 32U), (unsigned long)(uint32_t)addressCol);
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }

		j = addressLo + chunk;
		for (i = 0; i < rowBytes; i += radix) {
			switch (radix) {
				case 0x1U:
					val8 = *p8++;
					if (i >= addressLo && i < j) {
						nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)val8);
            if (nwritten > 0) {
              pPos += nwritten;
              remaining -= nwritten;
            }
					} else {
            if (remaining) {
  						*pPos++ = _T('.');
              remaining--;
            }
            if (remaining) {
  						*pPos++ = _T('.');
              remaining--;
            }
					}
					break;

				case 0x2U:
          if (bBigEndian) {
					  val16 = (((uint16_t)p8[0]) << 8U) | (((uint16_t)p8[1]) << 0U);
          } else {
					  val16 = (((uint16_t)p8[0]) << 0U) | (((uint16_t)p8[1]) << 8U);
          }
          p8 += 2;
					for (k = 2; k != 0;) {
						k--;
						if (i + k >= addressLo && i + k < j) {
							nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val16 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
						} else {
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
						}
						if (k == 3) {
              if (remaining) {
  							*pPos++ = _T('_');
                remaining--;
              }
						}
					}
					break;

				case 0x4U:
          if (bBigEndian) {
  					val32 = (((uint32_t)p8[0]) << 24U) | (((uint32_t)p8[1]) << 16U) | (((uint32_t)p8[2]) <<  8U) | (((uint32_t)p8[3]) <<  0U);
          } else {
	  				val32 = (((uint32_t)p8[0]) <<  0U) | (((uint32_t)p8[1]) <<  8U) | (((uint32_t)p8[2]) << 16U) | (((uint32_t)p8[3]) << 24U);
          }
          p8 += 4;
					for (k = 4; k != 0;) {
						k--;
						if (i + k >= addressLo && i + k < j) {
							nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val32 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
						} else {
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
						}
					}
					break;

				case 0x8U:
          if (bBigEndian) {
  					val64  = (((uint64_t)p8[0]) << 56U) | (((uint64_t)p8[1]) << 48U) | (((uint64_t)p8[2]) << 40U) | (((uint64_t)p8[3]) << 32U);
	  				val64 |= (((uint64_t)p8[4]) << 24U) | (((uint64_t)p8[5]) << 16U) | (((uint64_t)p8[6]) <<  8U) | (((uint64_t)p8[7]) <<  0U);
          } else {
  					val64  = (((uint64_t)p8[0]) <<  0U) | (((uint64_t)p8[1]) <<  8U) | (((uint64_t)p8[2]) << 16U) | (((uint64_t)p8[3]) << 24U);
	  				val64 |= (((uint64_t)p8[4]) << 32U) | (((uint64_t)p8[5]) << 40U) | (((uint64_t)p8[6]) << 48U) | (((uint64_t)p8[7]) << 56U);
          }
          p8 += 8;
					for (k = 8; k != 0;) {
						k--;
						if (i + k >= addressLo && i + k < j) {
							nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val64 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
						} else {
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
  						  *pPos++ = _T('.');
                remaining--;
              }
						}
						if (k == 4) {
              if (remaining) {
  							*pPos++ = _T('_');
                remaining--;
              }
						}
					}
					break;
			}
      if (remaining) {
			  *pPos++ = _T(' ');
        remaining--;
      }
		}

		p8 = (uint8_t*)columnBuf;
		for (i = 0; i < rowBytes; i++) {
			char c = (char)*p8++;

      if (remaining) {
			  if (_istprint(c) && c != _T('\t')) {
				  *pPos++ = c;
			  } else {
				  *pPos++ = _T('.');
			  }
        remaining--;
      }
		}
    if (remaining) {
      *pPos++ = _T('\n');
      remaining--;
    }
    if (remaining) {
      *pPos++ = _T('\0');
      remaining--;
    }
    line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
    _tprintf(_T("%s"), line);
		addressLo = 0;
		addressCol += rowBytes;
		m -= chunk;
		p += chunk;
    offset += chunk;
	}

  return TRUE; /* OK */
}

/* Returns TRUE if a token has just been fully parsed */
static bool_t
pumpTokenState(
  TokenState* pState,
  TCHAR c,
  bool_t bHex)
{
  TokenSubstate nextState = pState->substate;
  bool_t bParsed = FALSE, bStore = FALSE;

  if (c == _T('\0')) {
    c = _T(' ');
  } else {
    c = _totlower(c);
  }
  switch (pState->substate) {
  case NotInToken:
    if (!isspace(c)) {
      if (c >= _T('a') && c <= _T('f')) {
        nextState = InTokenHexadecimal;
        bStore = TRUE;
      } else {
        if (c >= _T('0') && c <= _T('9')) {
          nextState = bHex ? InTokenHexadecimal : InTokenDecimal;
          bStore = TRUE;
        } else {
          nextState = InTokenInvalid;
        }
      }
    }
    break;

  case InTokenDecimal:
    if (tolower(c) == _T('x') && pState->position == 1) {
      nextState = InTokenHexadecimal;
      pState->position = 0;
    } else if (c >= _T('a') && c <= _T('f')) {
      nextState = InTokenHexadecimal;
      bStore = TRUE;
    } else if (c >= _T('0') && c <= _T('9')) {
      nextState = InTokenDecimal;
      bStore = TRUE;
    } else if (isspace(c)) {
      nextState = NotInToken;
      bParsed = TRUE;
    } else {
      nextState = InTokenInvalid;
    }
    break;

  case InTokenHexadecimal:
    if ((c >= _T('a') && c <= _T('f')) || (c >= _T('0') && c <= _T('9'))) {
      nextState = InTokenHexadecimal;
      bStore = TRUE;
    } else if (c == _T('_')) {
      /* Ignore underscores inside hexadecimal tokens */
      nextState = InTokenHexadecimal;
    } else if (isspace(c)) {
      nextState = NotInToken;
      bParsed = TRUE;
    } else {
      pState->substate = InTokenInvalid;
    }
    break;

  case InTokenInvalid:
    if (isspace(c)) {
      nextState = NotInToken;
      bParsed = TRUE;
    } else {
      nextState = InTokenInvalid;
    }
    break;
  }

  if (bStore) {
    if (pState->position < ARRAY_LENGTH(pState->tokenBuf) - 1U) {
      pState->tokenBuf[pState->position] = c;
      pState->position++;
    }
  }

  if (bParsed) {
    pState->tokenBuf[pState->position] = _T('\0');
    if (pState->position == 0) {
      pState->value = 0;
    } else {
      if (pState->substate == InTokenHexadecimal) {
        _stscanf_s(pState->tokenBuf, _T("%llX"), &pState->value);
      } else if (pState->substate == InTokenDecimal) {
        _stscanf_s(pState->tokenBuf, _T("%llu"), &pState->value);
      } else {
        pState->value = 0;
      }
      pState->position = 0;
    }
  }
  pState->substate = nextState;

  return bParsed;
}

static bool_t
writeValueToWindowUnaligned(
  uint8_t* p,
  unsigned int width,
  bool_t bBigEndian,
  uint64_t val64)
{
  switch (width) {
  case 1U:
    *p++ = (uint8_t)val64;
    break;

  case 2U:
    if (bBigEndian) {
      *p++ = (uint8_t)(val64 >> 8U);
      *p++ = (uint8_t)(val64 >> 0U);
    } else {
      *p++ = (uint8_t)(val64 >> 0U);
      *p++ = (uint8_t)(val64 >> 8U);
    }
    break;

  case 4U:
    if (bBigEndian) {
      *p++ = (uint8_t)(val64 >> 24U);
      *p++ = (uint8_t)(val64 >> 16U);
      *p++ = (uint8_t)(val64 >> 8U);
      *p++ = (uint8_t)(val64 >> 0U);
    } else {
      *p++ = (uint8_t)(val64 >> 0U);
      *p++ = (uint8_t)(val64 >> 8U);
      *p++ = (uint8_t)(val64 >> 16U);
      *p++ = (uint8_t)(val64 >> 24U);
    }
    break;

  case 8U:
    if (bBigEndian) {
      *p++ = (uint8_t)(val64 >> 56U);
      *p++ = (uint8_t)(val64 >> 48U);
      *p++ = (uint8_t)(val64 >> 40U);
      *p++ = (uint8_t)(val64 >> 32U);
      *p++ = (uint8_t)(val64 >> 24U);
      *p++ = (uint8_t)(val64 >> 16U);
      *p++ = (uint8_t)(val64 >> 8U);
      *p++ = (uint8_t)(val64 >> 0U);
    } else {
      *p++ = (uint8_t)(val64 >> 0U);
      *p++ = (uint8_t)(val64 >> 8U);
      *p++ = (uint8_t)(val64 >> 16U);
      *p++ = (uint8_t)(val64 >> 24U);
      *p++ = (uint8_t)(val64 >> 32U);
      *p++ = (uint8_t)(val64 >> 40U);
      *p++ = (uint8_t)(val64 >> 48U);
      *p++ = (uint8_t)(val64 >> 56U);
    }
    break;

  default:
    assert(FALSE);
    break;
  }

  return TRUE; /* OK */
}

static bool_t
writeValueToWindow(
  uint8_t* p,
  unsigned int width,
  bool_t bBigEndian,
  uint64_t val64)
{
  uint16_t* p16;
  uint32_t* p32;
  uint64_t* p64;

  if ((uintptr_t)p & (width - 1U)) {
    return writeValueToWindowUnaligned(p, width, bBigEndian, val64);
  }

  switch (width) {
  case 1U:
    *p++ = (uint8_t)val64;
    break;

  case 2U:
    p16 = (uint16_t*)p;
    if (bBigEndian) {
      *p16 = (uint16_t)cpu_to_be16((uint16_t)val64);
    } else {
      *p16 = (uint16_t)cpu_to_le16((uint16_t)val64);
    }
    break;

  case 4U:
    p32 = (uint32_t*)p;
    if (bBigEndian) {
      *p32 = (uint32_t)cpu_to_be32((uint32_t)val64);
    } else {
      *p32 = (uint32_t)cpu_to_le32((uint32_t)val64);
    }
    break;

  case 8U:
    p64 = (uint64_t*)p;
    if (bBigEndian) {
      *p64 = (uint64_t)cpu_to_be64(val64);
    } else {
      *p64 = (uint64_t)cpu_to_le64(val64);
    }
    break;

  default:
    assert(FALSE);
    break;
  }

  return TRUE; /* OK */
}

int
_tmain(
  int argc,
  TCHAR* argv[])
{
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  uint8_t* pMapped = NULL;
  Option options[4];
  Arguments arguments = g_initArguments;
	OptionType paramType[5];
	TCHAR* paramInfo[5];
	TCHAR* paramMeaning[5];
  bool_t bBigEndian, bHex;
  CommandType command;
  unsigned int n, width, index, serial, windowIndex;
  uint64_t offset;
  TCHAR* pCommand;
  int ret = 0;

	options[0].key = _T("index");
	options[0].type = OptionUInt;
	options[0].def.uintVal = 0;
	options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
	options[2].key = _T("be");
	options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
	options[2].meaning = _T("TRUE => use big-endian memory format");
	options[3].key = _T("hex");
	options[3].type = OptionBoolean;
  options[3].def.booleanVal = FALSE;
	options[3].meaning = _T("TRUE => interpret values as being hexadecimal");
	arguments.nOption = 4;
	arguments.option = options;
	paramType[0] = OptionString;
	paramInfo[0] = _T("command");
  paramMeaning[0] = _T("Command to perform: fb, fw, fd, fq, rb, rw, rd, rq, wb, ww, wd or wq");
	paramType[1] = OptionUInt;
	paramInfo[1] = _T("window");
	paramMeaning[1] = _T("Index of window to access");
	paramType[2] = OptionUInt64;
	paramInfo[2] = _T("offset");
	paramMeaning[2] = _T("Offset in window");
	paramType[3] = OptionUInt;
	paramInfo[3] = _T("length");
	paramMeaning[3] = _T("Number of bytes to fill, read or write");
	paramType[4] = OptionString;
	paramInfo[4] = _T("item ...");
	paramMeaning[4] = _T("Fill value or data items to write");
	arguments.nParamType = 5;
	arguments.minNParam = 3;
	arguments.paramType = paramType;
	arguments.paramInfo = paramInfo;
	arguments.paramMeaning = paramMeaning;
  arguments.bNoStdOptions = TRUE;
	
  if (sampleParseCommandLine(_T("dump"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bBigEndian = options[2].value.booleanVal;
  bHex = options[3].value.booleanVal;
  pCommand = arguments.param[0].pStringVal;
  windowIndex = arguments.param[1].uintVal;
  offset = arguments.param[2].uint64Val;

  if (_tcsicmp(pCommand, _T("fb")) == 0) {
    width = 1;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fw")) == 0) {
    width = 2;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fd")) == 0) {
    width = 4;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fq")) == 0) {
    width = 8;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("rb")) == 0) {
    width = 1;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rw")) == 0) {
    width = 2;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rd")) == 0) {
    width = 4;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rq")) == 0) {
    width = 8;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("wb")) == 0) {
    width = 1;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("ww")) == 0) {
    width = 2;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("wd")) == 0) {
    width = 4;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("wq")) == 0) {
    width = 8;
    command = CommandTypeWrite;
  } else {
    _tprintf(_T("*** Invalid command: %s\n"), pCommand);
    ret = -1;
    goto done;
  }

  if (arguments.nParam < 4) {
    n = width;
  } else {
    n = arguments.param[3].uintVal;
  }
  if (n % width) {
    _tprintf(_T("+++ Warning: byte count %u(0x%X) is not an integer multiple of width (%u)\n"), n, n, width);
    n &= ~(width - 1U);
    _tprintf(_T("+++          byte count adjusted to %u(0x%X)\n"), n, n);
  }
  if (n == 0) {
    _tprintf(_T("+++ Warning: adjusted byte count is 0 (nothing to do).\n"));
    goto done;
  }

  status = sampleOpenCardNoClearError(index, options[0].specified, serial, options[1].specified, FALSE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_MapWindow(hCard, windowIndex, offset, n, (void**)&pMapped);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to map window %u offset 0x%llX + 0x%X: %s\n"),
      windowIndex, (unsigned long long)offset, n, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }
  _tprintf(_T("Window %u offset 0x%llX mapped @ %p\n"), windowIndex, (unsigned long long)offset, pMapped);

  if (command == CommandTypeFill) {
    TokenState state = g_tokenStateInit;
    TCHAR* pArg;
    TCHAR ch;
#if defined(_UNICODE)
    wint_t c;
#else
    int c;
#endif
    uint8_t* p = (uint8_t*)pMapped;
    size_t j;
    int i;
    uint64_t fillValue;

    i = 4;
    if (i < arguments.nParam) {
      /* Read token for fill value from remaining command line arguments */
      state = g_tokenStateInit;
      pArg = arguments.param[4].pStringVal;
      j = 0;
      do {
        ch = pArg[j];
        j++;
        if (pumpTokenState(&state, ch, bHex)) {
          break;
        }
      } while (_T('\0') != ch);
      fillValue = (uint64_t)state.value;
      _tprintf(_T("Fill value: 0x%llX\n"), (unsigned long long)fillValue);
    } else {
      /* Read token for fill value from stdin */
      _tprintf(_T("Provide fill value: "));
      state = g_tokenStateInit;
      do {
        c = _gettc(stdin);
        ch = (c == EOF_VALUE) ? _T('\0') : (TCHAR)c;
        if (pumpTokenState(&state, ch, bHex)) {
          break;
        }
      } while (c != EOF_VALUE);
      fillValue = (uint64_t)state.value;
    }
    if (!fillMemory(p, width, bBigEndian, fillValue, n)) {
      ret = -1;
      goto done;
    }
  } else if (command == CommandTypeWrite) {
    TokenState state = g_tokenStateInit;
    TCHAR* pArg;
    TCHAR ch;
#if defined(_UNICODE)
    wint_t c;
#else
    int c;
#endif
    uint8_t* p = (uint8_t*)pMapped;
    size_t j;
    int i;

    i = 4;
    while (i < arguments.nParam && n > 0) {
      /* Read token for next data item to write from remaining command line arguments */
      state = g_tokenStateInit;
      pArg = arguments.param[i].pStringVal;
      j = 0;
      do {
        ch = pArg[j];
        j++;
        if (pumpTokenState(&state, ch, bHex)) {
          _tprintf(_T("0x%08lX_%08lX: 0x%llX\n"),
            (unsigned long)(uint32_t)(offset >> 32), (unsigned long)(uint32_t)offset,
            state.value);
          if (!writeValueToWindow(p, width, bBigEndian, (uint64_t)state.value)) {
            ret = -1;
            break;
          }
          n -= width;
          offset += width;
          p += width;
        }
      } while (_T('\0') != ch);
      i++;
    }
    if (n == 0) {
      goto done;
    }

    /* Read tokens for data items from stdin to satisfy remaining byte count (if nonzero) */
    state = g_tokenStateInit;
    _tprintf(_T("Provide data for 0x%08lX_%08lX: "),
      (unsigned long)(uint32_t)(offset >> 32), (unsigned long)(uint32_t)offset);
    fflush(stdout);
    do {
      c = _gettc(stdin);
      ch = (c == EOF_VALUE) ? _T('\0') : (TCHAR)c;
      if (pumpTokenState(&state, ch, bHex)) {
        if (!writeValueToWindow(p, width, bBigEndian, (uint64_t)state.value)) {
          ret = -1;
          break;
        }
        n -= width;
        offset += width;
        p += width;
        if (n > 0) {
          _tprintf(_T("Provide data for 0x%08lX_%08lX: "),
            (unsigned long)(uint32_t)(offset >> 32), (unsigned long)(uint32_t)offset);
          fflush(stdout);
        }
      }
    } while (n > 0 && c != EOF_VALUE);
  } else if (command == CommandTypeRead) {
    if (!dumpMemoryViaWindow(0, 16 / width, width, bBigEndian, TRUE, offset, pMapped, n)) {
      ret = -1;
    }
  } else {
    /* Should never get here */
    assert(FALSE);
  }

done:
  if (NULL != pMapped) {
    status = ADMXRC3_UnmapWindow(hCard, pMapped);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("+++ Failed to unmap window at %p: %s\n"), pMapped, ADMXRC3_GetStatusString(status, TRUE));
    }
    pMapped = NULL;
  }

  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
