#pragma once

#include <platform.h>

class CTraceBuffer {
public:
  struct Point {
    timestamp_t m_timestamp;
    float m_value;
    bool m_bDiscontinuity;
  };

  CTraceBuffer(unsigned int sensorIndex) {
    m_lastTimestamp = TIMESTAMP_INVALID;
    m_pDataPoints = NULL;
    m_length = 0;
    m_head = 0;
    m_tail = 0;
    m_maxLength = 0;
    m_bMinMaxMeanValid = FALSE;
    m_sensorIndex = sensorIndex;
  }
  virtual ~CTraceBuffer() {
    if (NULL != m_pDataPoints) {
      delete[] m_pDataPoints;
      m_pDataPoints = NULL;
    }
  }

  void setSize(
    unsigned int size)
  {
    Point* pNewPoints;
    unsigned int from;

    ASSERT(size > 0);
    pNewPoints = new Point[size];
    if (size <= m_length) {
      ASSERT(size <= m_maxLength);
      from = moduloSub(m_tail, size, m_maxLength);
      m_sum = copyData(pNewPoints, from, size);
      m_length = size;
      m_head = 0;
      m_tail = 0;
    } else {
      ASSERT(m_length <= m_maxLength);
      from = moduloSub(m_tail, m_length, m_maxLength);
      m_sum = copyData(pNewPoints, from, m_length);
      m_head = 0;
      m_tail = m_length;
    }
    m_maxLength = size;
    delete[] m_pDataPoints;
    m_pDataPoints = pNewPoints;
  }

  void addPoint(const Point& point) {
    ASSERT(NULL != m_pDataPoints);
    ASSERT(m_maxLength > 0);
    ASSERT(m_length == 0 || point.m_timestamp > m_lastTimestamp);
    m_pDataPoints[m_tail] = point;
    m_lastTimestamp = point.m_timestamp;
    m_sum += point.m_value;
    if (!m_bMinMaxMeanValid) {
      m_min = m_max = point.m_value;
      m_bMinMaxMeanValid = TRUE;
    } else {
      if (point.m_value < m_min) {
        m_min = point.m_value;
      }
      if (point.m_value > m_max) {
        m_max = point.m_value;
      }
    }
    m_tail = moduloAdd(m_tail, 1, m_maxLength);
    if (m_length == m_maxLength) {
      m_sum -= m_pDataPoints[m_head].m_value;
      m_head = moduloAdd(m_head, 1, m_maxLength);
    } else {
      m_length++;
    }
  }

  unsigned int getLength() const {
    return m_length;
  }

  unsigned int getSize() const {
    return m_maxLength;
  }

  const Point& getPoint(unsigned int position) const {
    ASSERT(NULL != m_pDataPoints);
    ASSERT(position < m_length);

    position = moduloAdd(position, m_head, m_maxLength);
    return m_pDataPoints[position];
  }

  float getValue(unsigned int position) const {
    ASSERT(NULL != m_pDataPoints);
    ASSERT(position < m_length);

    position = moduloAdd(position, m_head, m_maxLength);
    return m_pDataPoints[position].m_value;
  }

  timestamp_t getTime(unsigned int position) const {
    ASSERT(NULL != m_pDataPoints);
    ASSERT(position < m_length);

    position = moduloAdd(position, m_head, m_maxLength);
    return m_pDataPoints[position].m_timestamp;
  }

  unsigned int getSensorIndex() const {
    return m_sensorIndex;
  }

  float getMin() const {
    ASSERT(m_bMinMaxMeanValid);
    return m_min;
  }

  float getMax() const {
    ASSERT(m_bMinMaxMeanValid);
    return m_max;
  }

  float getMean() const {
    ASSERT(m_bMinMaxMeanValid);
    return m_sum / (float)m_length;
  }

  timestamp_t getMinTime() const {
    ASSERT(m_length > 0);
    return m_pDataPoints[m_head].m_timestamp;
  }

  timestamp_t getMaxTime() const {
    ASSERT(m_length > 0);
    unsigned int position = moduloSub(m_tail, 1, m_maxLength);
    return m_pDataPoints[position].m_timestamp;
  }

  /* Find index of the point whose timestamp is greater than or equal to the specified timestamp */
  bool findTimeGE(timestamp_t time, unsigned int& index) {
    unsigned int index1, index2;
    if (findTimes(time, index1, index2)) {
      const Point& p1 = getPoint(index1);
      if (p1.m_timestamp == time) {
        index = index1;
      } else {
        index = index2;
      }
      return true;
    } else {
      return false;
    }
  }

  /* Find index of the point whose timestamp is greater than or equal to the specified timestamp */
  bool findTimeLE(timestamp_t time, unsigned int& index) {
    unsigned int index1, index2;
    if (findTimes(time, index1, index2)) {
      const Point& p2 = getPoint(index2);
      if (p2.m_timestamp == time) {
        index = index2;
      } else {
        index = index1;
      }
      return true;
    } else {
      return false;
    }
  }

  /* Find indices of points whose timestamp interval contains the requested timestamp */
  bool findTimes(timestamp_t time, unsigned int& index1, unsigned int& index2) {
    unsigned int i, j, k;
    timestamp_t ti, tj, tk;
    if (m_length < 2) {
      return false;
    } else {
      /* Binary search for pair of points that enclose specified timestamp */
      i = 0;
      j = m_length - 1;
      ti = getTime(i);
      tj = getTime(j);
      if (time < ti || time > tj) {
        /* Requested timestamp is outside this dataset */
        return false;
      }
      while (i + 1 != j) {
        k = (i + j) >> 1;
        tk = getTime(k);
        if (time <= tk) {
          j = k;
          tj = tk;
        } else {
          i = k;
          ti = k;
        }
      }
      if (getPoint(j).m_bDiscontinuity) {
        /* Requested timestamp is somewhere between points i and j, but j was not continuous w.r.t to i, so return false */
        return false;
      } else {
        /* Requested timestamp is somewhere between points i and j */
        index1 = i;
        index2 = j;
        return true;
      }
    }
  }

  /* Get value at specified timestamp, interpolating linearly */
  bool getValueAt(timestamp_t time, float& value) {
    if (m_length == 0) {
      return false;
    } else if (m_length == 1) {
      const Point& point = getPoint(0);
      if (time == point.m_timestamp) {
        value = point.m_value;
        return true;
      } else {
        return false;
      }
    } else {
      unsigned int i, j;
      if (findTimes(time, i, j)) {
        /* Linear interpolation */
        const Point& pi = getPoint(i);
        const Point& pj = getPoint(j);
        float p = (float)(time - pi.m_timestamp) / (float)(pj.m_timestamp - pi.m_timestamp);
        float q = 1.0f - p;
        value = q * pi.m_value + p * pj.m_value;
        return true;
      } else {
        return false;
      }
    }
  }

  void clear() {
    m_head = 0;
    m_tail = 0;
    m_length = 0;
    m_sum = 0.0f;
    m_bMinMaxMeanValid = FALSE;
  }

protected:
  unsigned int moduloSub(
    unsigned int a,
    unsigned int b,
    unsigned int m) const
  {
    return (a >= b) ? a - b : m + a - b;
  }

  unsigned int moduloAdd(
    unsigned int a,
    unsigned int b,
    unsigned int m) const
  {
    unsigned int sum = a + b;

    return (sum < m) ? sum : sum - m;
  }

  float copyData(
    Point* pBuffer,
    unsigned int from,
    unsigned int n)
  {
    Point* p = m_pDataPoints + from;
    float sum = 0.0f;

    ASSERT(n == 0 || NULL != m_pDataPoints);
    ASSERT(n <= m_length);

    while (n) {
      sum += p->m_value;
      *pBuffer++ = *p++;
      from++;
      if (from >= m_maxLength) {
        from -= m_maxLength;
        p = m_pDataPoints;
      }
      n--;
    }

    return sum;
  }

protected:
  unsigned int m_maxLength;
  unsigned int m_length;
  unsigned int m_head, m_tail;
  bool_t m_bMinMaxMeanValid;
  Point* m_pDataPoints;
  float m_min, m_max;
  float m_sum;
  timestamp_t m_lastTimestamp;
  unsigned int m_sensorIndex;
};
