/*
** common.c - Functions to support VxWorks SDK examples.
**
** (C) Copyright 2010-2013 Alpha Data
*/

#include <vxWorks.h>
#include <taskLib.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <admxrc3.h>
#include "sdk_common.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 9)
typedef TASK_ID ADATA_TASK_ID;
#else
typedef int ADATA_TASK_ID;
#endif

/* The environment variable used to construct a .BIT filename */
static const char g_admxrc3SdkVar[] = "ADMXRC3_SDK";

/* Strip leading and trailing whitespace */
static void
stripWhitespace(
  char* pString,
  size_t max) /* TO DO - currently ignored */
{
  char* p;
  char* q;
  char c;

  /* Strip leading whitespace */
  p = pString;
  q = pString;
  while (1) {
    c = *p;
    if (c == '\0') {
      return;
    }
    if (!isspace((int)c)) {
      break;
    }
    p++;
  }
  while (1) {
    c = *p++;
    if (c == '\0' || isspace((int)c)) {
      break;
    }
    *q++ = c;
  }
  *q = '\0';
}

/* Strip trailing slash from path */
static void
stripTrailingSlash(
  char* pString,
  size_t max)
{
  size_t n = strlen(pString);

  if (0 == n || n >= max + 1) {
    return;
  }
  if (pString[n - 1]  == '\\' || pString[n - 1] == '/') {
    pString[n - 1] = '\0';
  }
}

static void
stringLower(
  char* pDst,
  size_t max)
{
  char c;

  while (max) {
    c = *pDst;
    if ('\0' == c) {
      break;
    }
    *pDst++ = tolower(c);
  }
}

static void
stringCat(
  char* pDst,
  const char* pSrc,
  size_t max)
{
  size_t n = strlen(pDst);
  size_t m = strlen(pSrc);

  if (n >= max - 1) {
    return;
  }
  pDst += n;
  max -= n;
  if (m > max - 1) {
    m = max - 1;
  }
  if (m) {
    memcpy(pDst, pSrc, m * sizeof(char));
    pDst[m] = '\0';
  }
}

ADMXRC3_STATUS
sampleGetFpgaFilename(
  ADMXRC3_HANDLE hCard,
  unsigned int targetIndex,
  const char* pDir,
  const char* pDesign,
  size_t bufferSize,
  char* pFileNameBuffer,
  bool_t* pbShouldConfigure)
{
  ADMXRC3_CARD_INFO cardInfo;
  ADMXRC3_FPGA_INFO fpgaInfo;
  ADMXRC3_STATUS status;
  const char* pEnvValue;
  const char* pModelName;
  bool_t bShouldConfigure = TRUE;
  
  status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
  if (ADMXRC3_SUCCESS != status) {
    return status;
  }

  switch(cardInfo.Model)
  {
  case ADMXRC3_MODEL_ADMXRC2:
    pModelName = "admxrc2";
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
    pModelName = "adpexrc5t";
    break;

  case ADMXRC3_MODEL_ADMXRC6TL:
    pModelName = "admxrc6tl";
    break;

  case ADMXRC3_MODEL_ADMXRC6T1:
    pModelName = "admxrc6t1";
    break;

  case ADMXRC3_MODEL_ADMXRC6TGE:
    pModelName = "admxrc6tge";
    break;

  case ADMXRC3_MODEL_ADMXRC6TGEL:
    pModelName = "admxrc6tgel";
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
    bShouldConfigure = FALSE;
    pModelName = "admxrc6tadv8";
    break;

  case ADMXRC3_MODEL_ADPEXRC6T:
    pModelName = "adpexrc6t";
    break;

  case ADMXRC3_MODEL_ADPEXRC6TL:
    pModelName = "adpexrc6tl";
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    pModelName = "adpexrc6tadv";
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_T:
    /* Target FPGA is not configurable via ADMXRC3 API */
    bShouldConfigure = FALSE;
    pModelName = "adpexrc6tadv";
    break;

  case ADMXRC3_MODEL_ADMXRC7K1:
    pModelName = "admxrc7k1";
    break;

  case ADMXRC3_MODEL_ADMXRC6TDA1:
    pModelName = "admxrc6tda1";
    break;

  case ADMXRC3_MODEL_ADMXRC7V1:
    pModelName = "admxrc7v1";
    break;

  case ADMXRC3_MODEL_ADMVPX37V2:
    pModelName = "admvpx37v2";
    break;

  case ADMXRC3_MODEL_ADMPCIE7V3:
    /* Target FPGA is not configurable via ADMXRC3 API */
    bShouldConfigure = FALSE;
    pModelName = "admpcie7v3";
    break;

  case ADMXRC3_MODEL_GENERIC:
  default:
    return ADMXRC3_NOT_SUPPORTED;
  }

  status = ADMXRC3_GetFpgaInfo(hCard, targetIndex, &fpgaInfo);
  if (ADMXRC3_SUCCESS != status) {
    return status;
  }
  stringLower(fpgaInfo.Device, ARRAY_LENGTH(fpgaInfo.Device));

  pEnvValue = getenv(g_admxrc3SdkVar);
  if (NULL == pEnvValue) {
    pEnvValue = "";
  }

  if (bufferSize) {
    pFileNameBuffer[0] = '\0';
  }
  stringCat(pFileNameBuffer, pEnvValue, bufferSize);
  stripWhitespace(pFileNameBuffer, bufferSize);
  stripTrailingSlash(pFileNameBuffer, bufferSize);
  stringCat(pFileNameBuffer, "/bit/", bufferSize);
  stringCat(pFileNameBuffer, pDir, bufferSize);
  stringCat(pFileNameBuffer, "/", bufferSize);
  stringCat(pFileNameBuffer, pDesign, bufferSize);
  stringCat(pFileNameBuffer, "-", bufferSize);
  stringCat(pFileNameBuffer, pModelName, bufferSize);
  stringCat(pFileNameBuffer, "-", bufferSize);
  stringCat(pFileNameBuffer, fpgaInfo.Device, bufferSize);
  stringCat(pFileNameBuffer, ".bit", bufferSize);

  if (NULL != pbShouldConfigure) {
    *pbShouldConfigure = bShouldConfigure;
  }
  return status;
}

ADMXRC3_STATUS
sampleOpenCard(
  unsigned int index,
  bool_t bByIndex,
  unsigned int serial,
  bool_t bBySerial,
  ADMXRC3_HANDLE* phCard)
{
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hCard;
  ADMXRC3_CARD_INFO cardInfo;
  unsigned int i;

  if (bByIndex && bBySerial) {
    /* Attempt to open card with specified index and serial number */
    status = ADMXRC3_Open(index, &hCard);
    if (status != ADMXRC3_SUCCESS) {
      return status;
    }
    status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
    if (status != ADMXRC3_SUCCESS) {
      return status;
    }
    if (cardInfo.SerialNumber == serial) {
      *phCard = hCard;
      return ADMXRC3_SUCCESS;
    }
    return ADMXRC3_DEVICE_NOT_FOUND;
  }

  if (bBySerial) {
    /* Attempt to open card with specified serial number */
    for (i = 0; i < 256; i++) {
      status = ADMXRC3_Open(i, &hCard);
      if (status != ADMXRC3_SUCCESS) {
        if (status == ADMXRC3_DEVICE_NOT_FOUND) {
          return status;
        }
        continue;
      }
      status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
      if (status != ADMXRC3_SUCCESS) {
        continue;
      }
      if (cardInfo.SerialNumber == serial) {
        *phCard = hCard;
        return ADMXRC3_SUCCESS;
      }
      ADMXRC3_Close(hCard);
    }
    return ADMXRC3_DEVICE_NOT_FOUND;
  } else {
    /* Attempt to open card with specified index */
    return ADMXRC3_Open(index, phCard);
  }
}

bool_t
sampleTaskNice(
  void)
{
  ADATA_TASK_ID tId = taskIdSelf();
  int priority;

  if (OK == taskPriorityGet(tId, &priority)) {
    if (priority < 255) {
      if (OK == taskPrioritySet(tId, 255)) {
        return TRUE;
      } else {
        fprintf(stderr, "*** sampleTaskNice: failed to set our priority, errno=%d(0x%x)\n", errno, errno);
        return FALSE;
      }
    }
  } else {
    fprintf(stderr, "*** sampleTaskNice: failed to get our priority, errno=%d(0x%x)\n", errno, errno);
    return FALSE;
  }
  return TRUE;
}
