/*
** flash.c - utility to program, erase and verify the Flash memory on
**           an ADM-XRC Gen 3 card.
**
** Warning: Although this utility performs checks in order to ensure
**          that you do not attempt to program the Flash memory with a
**          bitstream for the wrong type of FPGA, ensure that you do
**          not program the flash with an invalid bitstream. Doing so
**          and powering up the FPGA for an extended period of time
**          can damage devices on the card.
**
** Utility for:
**
**   o ADPe-XRC-5T
**   o ADM-XRC-6TL
**   o ADM-XRC-6T1
**   o ADM-XRC-6T-DA1
**   o ADM-XRC-6TGE & ADM-XRC-6TGEL
**   o ADM-XRC-6T-ADV8
**   o ADPE-XRC-6T & ADPE-XRC-6T-L
**   o ADPE-XRC-6T-ADV (Controller)
**   o ADM-XRC-7K1
**   o ADM-XRC-7V1
**   o ADM-VPX3-7V2
**
** (C) Copyright Alpha Data 2009-2013
**
** NOTE: This utility applies only to the regions marked (*) in the comments
**       below, and cannot affect the other regions.
**
**  1. The Flash address map used on the ADM-XRC-6TL, ADM-XRC-6T1,
**     ADM-XRC-6T-DA1, ADM-XRC-6TGE(L), ADPE-XRC-6T(-L), ADM-XRC-7K1 &
**     ADM-XRC-7V1 is:
**
**     Flash bank 0:
**     0x0000000 - 0x07FFFFF  Alternate ADB3 PCI-E to OCP Bridge bitstream
**     0x0800000 - 0x0FFFFFF  Normal ADB3 PCI-E to OCP Bridge bitstream
**     0x1000000 - 0x10FFFFF  Alpha Data VPD region
**     0x1100000 - 0x11FFFFF  Customer VPD region
**     0x1200000 - 0x28FFFFF  Target FPGA bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA bitstream (*)
**
**  2. The Flash address map used on the ADM-XRC-6T-ADV8 is:
**
**     Flash bank 0:
**     0x0000000 - 0x1FFFFFF  Failsafe FPGA bitstream (*)
**     0x2000000 - 0x3FFFFFF  Normal FPGA bitstream (*)
**
**  3. The Flash address map used on the ADM-XRC-6T-ADV (Controller) is:
**
**     Flash bank 0 (U52):
**     0x0000000 - 0x07FFFFF  Alternate Controller FPGA bitstream
**     0x0800000 - 0x0FFFFFF  Normal Controller FPGA bitstream
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 0 bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 0 bitstream (*)
**
**     Flash bank 1 (U51):
**     0x0000000 - 0x0FFFFFF  Customer region
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 1 bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 1 bitstream (*)
**
**  4. The Flash address map used on the ADM-VPX3-7V2 is:
**
**     Flash bank 0 (U6) when size of bank is 64 MiB:
**     0x0000000 - 0x07FFFFF  Alternate Controller FPGA bitstream
**     0x0800000 - 0x0FFFFFF  Normal Controller FPGA bitstream
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 0 bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 0 bitstream (*)
**
**     Flash bank 0 (U6) when size of bank is 128 MiB:
**     0x0000000 - 0x07FFFFF  Alternate Controller FPGA bitstream
**     0x0800000 - 0x0FFFFFF  Normal Controller FPGA bitstream
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x48FFFFF  Target FPGA 0 bitstream (*)
**     0x4900000 - 0x7FFFFFF  Failsafe target FPGA 0 bitstream (*)
**
**     NOTE: The above address maps are logical, i.e. from the point of view of
**     somebody calling the Flash functions in the ADMXRC3 API. In order to
**     support both 64 MiB and 128 MiB devices without making regions
**     noncontiguous, the Alpha Data ADB3 Driver remaps the physical layout to
**     the above logical layout. The phyiscal layout is as follows:
**
**     Flash bank 0 (U6):
**     0x0000000 - 0x07FFFFF  Alternate Controller FPGA bitstream
**     0x0800000 - 0x0FFFFFF  Normal Controller FPGA bitstream
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 0 bitstream, lower 23 MiB (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 0 bitstream, lower 23 MiB (*)
**     (Next two regions present only in a 128 MiB bank)
**     0x4000000 - 0x5FFFFFF  Target FPGA 0 bitstream, upper 32 MiB (*)
**     0x6000000 - 0x7FFFFFF  Failsafe target FPGA 0 bitstream, upper 32 MiB (*)
*/

#include <vxWorks.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <admxrc3.h>
#include "sdk_common.h"
#include "platform_endian.h"

#ifndef FALSE
# define FALSE (0)
#endif
#ifndef TRUE
# define TRUE (1)
#endif

#define BLANK_CHECK_BUFFER_SIZE (0x10000)
#define READBACK_BUFFER_SIZE (0x10000)
#define VERIFY_BUFFER_SIZE (0x10000)

typedef enum _BootFlagType {
  /* ADM-XRC-6T-ADV8 */
  BootFlagTypeNone,          /* No boot flag is programmed into Flash */

  /* ADM-XRC-II */
  BootFlagType8Bit,          /* [7:0] = 0 */

  /*
  ** ADPe-XRC-5T, ADM-XRC-6TL, ADM-XRC-6T1, ADM-XRC-6T-DA1, ADM-XRC-6TGE(L),
  ** ADM-XRC-6T-ADV8, ADPE-XRC-6T(-L), ADPE-XRC-6T-ADV, ADM-XRC-7K1,
  ** ADM-XRC-7V1
  */
  BootFlagType24BitAddress,  /* [31:8] = length in bytes - 1, [7:0] = 0 */

  /* ADM-VPX3-7V2 */
  BootFlagType30BitAddress   /* [31] = 0, [30] = (reserved), [29:0] = length in bytes - 1 */
} BootFlagType;

static const char g_funcName[] = "admxrc3Flash";

static int
ourStrCaseCmp(
  const char* pS1,
  const char* pS2)
{
  char c1, c2;

  while (1) {
    c1 = *pS1;
    c2 = *pS2;
    if (c1 == '\0' || tolower(c1) != tolower(c2)) {
      break;
    }
    pS1++;
    pS2++;
  }
  return tolower(c1) - tolower(c2);
}

static void
showUsage(
  const char* pFuncName)
{
  printf("Usage:\n");
  printf("    %s <index>, <flags>, \"info\"\n", pFuncName);
  printf("    %s <index>, <flags>, \"chkblank\", <target index>\n", pFuncName);
  printf("    %s <index>, <flags>, \"erase\",    <target index>\n", pFuncName);
  printf("    %s <index>, <flags>, \"program\",  <target index>, <.BIT filename>\n", pFuncName);
  printf("    %s <index>, <flags>, \"verify\",   <target index>, <.BIT filename>\n", pFuncName);
  printf("where <flags> is one or more of:\n");
  printf("    0x001 => BYSERIAL  interpret <index> as serial number of card to open\n");
  printf("    0x010 => FORCE     ignore target FPGA / bitstream mismatch\n");
  printf("    0x100 => FAILSAFE  program or verify the failsafe image");
}

static ADMXRC3_STATUS
eraseFlash(
  ADMXRC3_HANDLE hCard,
  uint64_t eraseStart,
  uint64_t eraseLimit)
{
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint64_t address;

  address = eraseStart;
  while (address < eraseLimit) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to get Flash block info for location 0x%llX: %s\n", (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    printf("0x%llX\r", (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_EraseFlash(hCard, 0, 0, blockInfo.Address, blockInfo.Length);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to erase Flash at 0x%llX-0x%llX: %s\n",
        (unsigned long long)blockInfo.Address, (unsigned long long)(blockInfo.Address + blockInfo.Length - 1U), ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    /* Advance to next block */
    address = blockInfo.Address + blockInfo.Length;
    if (address <= eraseStart) {
      /* Catch 'address' wrapping around to 0 */
      break;
    }
  }

  return status;
}

static ADMXRC3_STATUS
programBitstream(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_BITSTREAM* pBitstream,
  uint64_t programStart)
{
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint8_t* pData;
  uint64_t address, chunk;
  size_t remaining;
  
  address = programStart;
  remaining = pBitstream->Length;
  pData = &pBitstream->Data[0];
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to get Flash block info for location 0x%llX: %s\n",
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > remaining) {
      chunk = remaining;
    }
    printf("0x%llX\r", (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_WriteFlash(hCard, 0, 0, address, (size_t)chunk, pData); /* Cast is safe because we guarantee chunk <= maximum size_t value */
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to write Flash at 0x%llX-0x%llX: %s\n",
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      break;
    }

    address += chunk;
    remaining -= (size_t)chunk; /* Cast is safe because we guarantee chunk <= maximum size_t value */
    pData += (size_t)chunk; /* Cast is safe because we guarantee chunk <= maximum size_t value */
  }

  return status;
}

static ADMXRC3_STATUS
checkBlank(
  ADMXRC3_HANDLE hCard,
  uint64_t checkStart,
  uint64_t checkLimit,
  uint64_t* pnError)
{
  uint8_t* pBuffer = NULL;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, remaining, nError, i, chunk;

  pBuffer = (uint8_t*)malloc(BLANK_CHECK_BUFFER_SIZE);
  if (NULL == pBuffer) {
    printf("*** Failed to allocate blank check buffer of size 0x%lX bytes\n",
      (unsigned long)BLANK_CHECK_BUFFER_SIZE);
    return ADMXRC3_NO_MEMORY;
  }

  nError = 0;
  address = checkStart;
  remaining = checkLimit - checkStart;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to get Flash block info for location 0x%llX: %s\n", (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > BLANK_CHECK_BUFFER_SIZE) {
      chunk = BLANK_CHECK_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    printf("0x%llX\r", (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hCard, 0, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to read back Flash memory at 0x%llX-0x%llX: %s\n",
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    for (i = 0; i < chunk; i++) {
      if (pBuffer[i] != 0xffU) {
        if (nError < 10) {
          printf("*** Error at Flash address 0x%llX: expected=0x%02X, actual=0x%02X\n",
            (unsigned long long)(address + i), 0xffU, pBuffer[i]);
        } else if (nError == 10) {
          printf("*** (more errors)\n");
        }
        nError++;
      }
    }

    address += chunk;
    remaining -= chunk;
  }

  free(pBuffer);

  *pnError = nError;
  return status;
}

static ADMXRC3_STATUS
verify(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_BITSTREAM* pBitstream,
  uint64_t verifyStart,
  uint64_t* pnError)
{
  uint8_t* pBuffer = NULL;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, chunk, nError;
  size_t i, offset, remaining;

  pBuffer = (uint8_t*)malloc(VERIFY_BUFFER_SIZE);
  if (NULL == pBuffer) {
    printf("*** Failed to allocate verify buffer of size 0x%lX bytes\n",
      (unsigned long)VERIFY_BUFFER_SIZE);
    return ADMXRC3_NO_MEMORY;
  }

  nError = 0;
  offset = 0;
  address = verifyStart;
  remaining = pBitstream->Length;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to get Flash block info for location 0x%llX: %s\n",
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > VERIFY_BUFFER_SIZE) {
      chunk = VERIFY_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    printf("0x%llX\r", (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hCard, 0, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to read back Flash memory at 0x%llX-0x%llX: %s\n",
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    for (i = 0; i < chunk; i++) {
      if (pBuffer[i] != pBitstream->Data[offset + i]) {
        if (nError < 10) {
          printf("*** Error at Flash address 0x%llX: expected=0x%02X, actual=0x%02X\n",
            (unsigned long long)(address + i), pBitstream->Data[offset + i], pBuffer[i]);
        } else if (nError == 10) {
          printf("*** (more errors)\n");
        }
        nError++;
      }
    }

    address += chunk;
    offset += (size_t)chunk; /* Cast is safe because we guarantee chunk < size of buffer */
    remaining -= (size_t)chunk; /* Cast is safe because we guarantee chunk < size of buffer */
  }

  free(pBuffer);

  *pnError = nError;
  return status;
}

static ADMXRC3_STATUS
readback(
  ADMXRC3_HANDLE hCard,
  uint64_t readAddress,
  uint64_t readLength,
  FILE* pFile)
{
  uint8_t* pBuffer;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, remaining, chunk;
  size_t n;

  pBuffer = (uint8_t*)malloc(READBACK_BUFFER_SIZE);
  if (NULL == pBuffer) {
    printf("*** Failed to allocate verify buffer of size 0x%lX bytes\n",
      (unsigned long)READBACK_BUFFER_SIZE);
    return ADMXRC3_NO_MEMORY;
  }

  address = readAddress;
  remaining = readLength;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to get Flash block info for location 0x%llX: %s\n",
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > READBACK_BUFFER_SIZE) {
      chunk = READBACK_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    printf("0x%llX\r", (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hCard, 0, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to read back Flash memory at 0x%llX-0x%llX: %s\n",
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    n = fwrite(pBuffer, 1, (size_t)chunk, pFile); /* Cast is safe because we guarantee chunk < size of buffer */
    if (n != chunk) {
      printf("*** Failed to write all data to file, %lu(0x%lX) actual vs. %lu(0x%lX) expected.\n",
        (unsigned long)n, (unsigned long)n, (unsigned long)chunk, (unsigned long)chunk);
      free(pBuffer);
      return ADMXRC3_UNEXPECTED_ERROR;
    }

    address += chunk;
    remaining -= chunk;
  }

  free(pBuffer);

  return status;
}

/* Bits for 'flags' parameter */
#define FLAG_BYSERIAL (0x1U)   /* Interpret 'index' as a serial number */
#define FLAG_FORCE (0x10U)     /* Do operation even if target FPGA device is wrong for specified .BIT file */
#define FLAG_FAILSAFE (0x100U) /* Program the failsafe image, rather than the default image */

STATUS
admxrc3Flash(
  unsigned int index,          /* Index of device to open (OR serial number if 'flags' includes BYSERIAL) */
  unsigned int flags,          /* See above bits */
  const char* pCommand,        /* "program", "verify", "erase", "chkblank" or "readback" */
  unsigned int targetIndex,    /* Index of target FPGA whose image in Flash is to be programmed */
  const char* pBitFileName,    /* Bitstream filename */
  size_t readbackOffset,       /* (Valid only for readback) Address at which to begin reading back */
  size_t readbackLength)       /* (Valid only for readback) Number of bytes to read back */
{
  STATUS ret = OK;
  const double clockFreq = 33.0e6;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO	cardInfo;
  ADMXRC3_FLASH_INFO flashInfo;
  ADMXRC3_FPGA_INFO fpgaInfo;
  ADMXRC3_BITSTREAM* pBitstream = NULL;
  unsigned int flashIndex = (unsigned int)-1;
  bool_t bShowInfo = FALSE, bDoCheckBlank = FALSE, bDoErase = FALSE, bDoProgram = FALSE, bDoReadback = FALSE, bDoVerify = FALSE;
  uint64_t bootFlagAddress = 0;
  uint64_t programAddress = 0;
  uint64_t eraseAddress = 0, eraseLimit  = 0;
  uint64_t nError = 0;
  bool_t bFpgaInfoValid = FALSE;
  BootFlagType bootFlagType = BootFlagTypeNone;
  bool_t bBootFlagVerify = TRUE;
  bool_t bBootFlagOk = FALSE;
  bool_t bFailsafeImage, bForce, bOpenBySerial;
  bool_t bBitSwap = FALSE;
  unsigned int byteSwap = 1;

  sampleTaskNice();

  if (pCommand == NULL) {
    printf("*** Please specify a valid command.\n");
    showUsage(g_funcName);
    ret = ERROR;
    goto done;
  }

  bOpenBySerial = (flags & FLAG_BYSERIAL) ? TRUE : FALSE;
  bForce = (flags & FLAG_FORCE) ? TRUE : FALSE;
  bFailsafeImage = (flags & FLAG_FAILSAFE) ? TRUE : FALSE;

  bShowInfo = ourStrCaseCmp(pCommand, "info") == 0;
  bDoProgram = ourStrCaseCmp(pCommand, "program") == 0;
  bDoVerify = ourStrCaseCmp(pCommand, "verify") == 0;
  bDoErase = ourStrCaseCmp(pCommand, "erase") == 0;
  bDoCheckBlank = ourStrCaseCmp(pCommand, "chkblank") == 0;
  bDoReadback = ourStrCaseCmp(pCommand, "readback") == 0;

  if (!bShowInfo && !bDoProgram && !bDoVerify && !bDoErase && !bDoCheckBlank && !bDoReadback) {
    printf("*** Please specify a valid command.\n");
    showUsage(g_funcName);
    ret = ERROR;
    goto done;
  }

  if (bDoProgram || bDoVerify) {
    if (NULL == pBitFileName) {
      printf("*** Must specify a filename for programming or verification\n");
      showUsage(g_funcName);
      ret = ERROR;
      goto done;
    }
  }

  if (bDoReadback) {
    if (NULL == pBitFileName || readbackLength == 0) {
      printf("*** Must specify a filename, offset and length for readback\n");
      ret = ERROR;
      goto done;
    }
  }

  status = sampleOpenCard(index, !bOpenBySerial, index, bOpenBySerial, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index/serial %lu: %s\n",
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRC2:
    if (bFailsafeImage) {
      printf("*** +failsafe option is not supported for this model.\n");
      ret = ERROR;
      goto done;
    }
    if (0 != targetIndex) {
      printf("*** Value %lu for 'target' option is invalid for this model.\n", (unsigned long)targetIndex);
      ret = ERROR;
      goto done;
    }
    flashIndex = 0U;
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADMXRC6TADV8:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC7V1:
  case ADMXRC3_MODEL_ADMVPX37V2:
  case ADMXRC3_MODEL_ADMPCIE7V3:
    if (0 != targetIndex) {
      printf("*** Value %lu for 'target' option is invalid for this model.\n", (unsigned long)targetIndex);
      ret = ERROR;
      goto done;
    }
    flashIndex = 0U;
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    if (targetIndex >= 2) {
      printf("*** Value %lu for 'target' option is invalid for this model.\n", (unsigned long)targetIndex);
      ret = ERROR;
      goto done;
    }
    flashIndex = targetIndex;
    break;

  default:
    printf("*** This utility must be run on one of the following cards:\n\n");
    printf("\tADM-XRC-II\n");
    printf("\tADPe-XRC-5T\n");
    printf("\tADM-XRC-6TL\n");
    printf("\tADM-XRC-6T1\n");
    printf("\tADM-XRC-6T-DA1\n");
    printf("\tADM-XRC-6TGE / ADM-XRC-6TGEL\n");
    printf("\tADM-XRC-6T-ADV8\n");
    printf("\tADPE-XRC-6T / ADPE-XRC-6T-L\n");
    printf("\tADPE-XRC-6T-ADV (Controller)\n");
    printf("\tADM-XRC-7K1\n");
    printf("\tADM-XRC-7V1\n");
    printf("\tADM-VPX3-7V2\n");
    printf("\tADM-PCIE-7V3\n");
    ret = ERROR;
    goto done;
  }

  status = ADMXRC3_GetFpgaInfo(hCard, targetIndex, &fpgaInfo);
  if (status != ADMXRC3_SUCCESS) {
    if (!bForce) {
      printf("*** Failed to get FPGA %lu info: %s\n", (unsigned long)targetIndex, ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    } else {
      printf("+++ WARNING: Failed to get FPGA %lu info: %s\n", (unsigned long)targetIndex, ADMXRC3_GetStatusString(status, TRUE));
      printf("             Overriding due to +force option specified on command line.\n");
    }
  } else {
    printf("Target FPGA is device %s\n", fpgaInfo.Identifier);
    bFpgaInfoValid = TRUE;
  }

  status = ADMXRC3_GetFlashInfo(hCard, 0, &flashInfo);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get Flash information: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRC2:
    bootFlagAddress = flashInfo.UseableStart + 0x8000U;
    programAddress = flashInfo.UseableStart + 0x8001U;
    eraseAddress = flashInfo.UseableStart;
    eraseLimit = flashInfo.UseableStart + flashInfo.UseableLength;
    bootFlagType = BootFlagType8Bit;
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
    if (bFailsafeImage) {
      bootFlagAddress = flashInfo.UseableStart + 0xC00000U;
      programAddress = flashInfo.UseableStart + 0xC20000U;
      eraseAddress = flashInfo.UseableStart + 0xC00000U;
      eraseLimit = flashInfo.UseableStart + 0x1800000U;
    } else {
      bootFlagAddress = flashInfo.UseableStart;
      programAddress = flashInfo.UseableStart + 0x20000U;
      eraseAddress = flashInfo.UseableStart;
      eraseLimit = flashInfo.UseableStart + 0xC00000U;
    }
    bootFlagType = BootFlagType24BitAddress;
    break;

  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC7V1:
    if (bFailsafeImage) {
      bootFlagAddress = flashInfo.UseableStart + 0x1700000U;
      programAddress = flashInfo.UseableStart + 0x1720000U;
      eraseAddress = flashInfo.UseableStart + 0x1700000U;
      eraseLimit = flashInfo.UseableStart + 0x2E00000U;
    } else {
      bootFlagAddress = flashInfo.UseableStart;
      programAddress = flashInfo.UseableStart + 0x20000U;
      eraseAddress = flashInfo.UseableStart;
      eraseLimit = flashInfo.UseableStart + 0x1700000U;
    }
    bootFlagType = BootFlagType24BitAddress;
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
    bBitSwap = TRUE;
    byteSwap = 2;
    if (bFailsafeImage) {
      programAddress = flashInfo.UseableStart + 0x0U;
      eraseAddress = flashInfo.UseableStart + 0x0U;
      eraseLimit = flashInfo.UseableStart + 0x2000000U;
    } else {
      programAddress = flashInfo.UseableStart + 0x2000000U;
      eraseAddress = flashInfo.UseableStart + 0x2000000U;
      eraseLimit = flashInfo.UseableStart + 0x4000000U;
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    if (bFailsafeImage) {
      bootFlagAddress = flashInfo.UseableStart + 0x1700000U;
      programAddress = flashInfo.UseableStart + 0x1720000U;
      eraseAddress = flashInfo.UseableStart + 0x1700000U;
      eraseLimit = flashInfo.UseableStart + 0x2E00000U;
    } else {
      bootFlagAddress = flashInfo.UseableStart;
      programAddress = flashInfo.UseableStart + 0x20000U;
      eraseAddress = flashInfo.UseableStart;
      eraseLimit = flashInfo.UseableStart + 0x1700000U;
    }
    bootFlagType = BootFlagType24BitAddress;
    break;

  case ADMXRC3_MODEL_ADMVPX37V2:
    /* The driver remaps the physical Flash layout so that the normal and
    ** failsafe images are exactly equal in size and contiguous, regardless
    ** of the size of Flash device fitted. */
    if (bFailsafeImage) {
      bootFlagAddress = flashInfo.UseableStart + (flashInfo.UseableLength / 2);
      programAddress = bootFlagAddress + 0x20000U;
      eraseAddress = bootFlagAddress;
      eraseLimit = bootFlagAddress + (flashInfo.UseableLength / 2);
    } else {
      bootFlagAddress = flashInfo.UseableStart;
      programAddress = bootFlagAddress + 0x20000U;
      eraseAddress = bootFlagAddress;
      eraseLimit = bootFlagAddress + (flashInfo.UseableLength / 2);
    }
    bootFlagType = BootFlagType30BitAddress;
    break;

  case ADMXRC3_MODEL_ADMPCIE7V3:
    bBitSwap = TRUE;
    byteSwap = 2;
    if (bFailsafeImage) {
      programAddress = flashInfo.UseableStart + 0x0U;
      eraseAddress = flashInfo.UseableStart + 0x0U;
      eraseLimit = flashInfo.UseableStart + 0x2000000U;
    } else {
      programAddress = flashInfo.UseableStart + 0x2000000U;
      eraseAddress = flashInfo.UseableStart + 0x2000000U;
      eraseLimit = flashInfo.UseableStart + 0x4000000U;
    }
    break;

  default:
    printf("*** Value %lu for 'target' option is invalid for this model.\n", (unsigned long)targetIndex);
    ret = ERROR;
    goto done;
  }

  printf("Flash type is %s, %llu(0x%llX) kiB\n", flashInfo.Identifier,
    (unsigned long long)(flashInfo.Size / 1024), (unsigned long long)(flashInfo.Size / 1024));
  printf("Useable region is 0x%llX-0x%llX\n",
    (unsigned long long)flashInfo.UseableStart, (unsigned long long)(flashInfo.UseableStart + flashInfo.UseableLength - 1U));

  if (bShowInfo) {
    goto done;
  }

  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    status = ADMXRC3_SetClockFrequency(hCard, ADMXRC3_CLOCK_LCLK, 0, clockFreq, NULL);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to set LCLK to %.1fMHz: %s\n", clockFreq / 1.0e6, ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
  }

  if (bDoProgram || bDoVerify) {
    uint32_t i;
    uint8_t val8;
    uint8_t val8a, val8b, val8c, val8d;

    /* Load the .BIT file specified on the command line */
    status = ADMXRC3_LoadBitstream(pBitFileName, &pBitstream);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to load bitstream '%s': %s\n", pBitFileName, ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
    printf("Bitstream is for device %s, length %lu(0x%lX) bytes\n",
      pBitstream->Identifier, (unsigned long)pBitstream->Length, (unsigned long)pBitstream->Length);
    if (bFpgaInfoValid) {
      if (ourStrCaseCmp(pBitstream->Identifier, fpgaInfo.Identifier) != 0) {
        if (!bForce) {
          printf("*** ERROR: Bitstream device does not match target FPGA device - aborting\n");
          ret = ERROR;
          goto done;
        } else {
          printf("WARNING: Device information in bitstream device does not match target FPGA device.\n");
          if (bDoProgram) {
            printf("         Overriding due to +force option specified on command line.\n");
          }
        }
      }
    } else {
      assert(bForce);
      printf("WARNING: Information for target FPGA was not found. Bitstream may or may not match actual target FPGA device.\n");
      printf("         Overriding due to +force option specified on command line.\n");
    }
    if (bBitSwap) {
      printf("Doing bit-swapping...\n");
      for (i = 0; i < pBitstream->Length; i++) {
        val8 = pBitstream->Data[i];
        pBitstream->Data[i] = (uint8_t)(((val8 & 0x80U) >> 7) | ((val8 & 0x40U) >> 5) | ((val8 & 0x20U) >> 3) | ((val8 & 0x10U) >> 1) |
                                        ((val8 & 0x08U) << 1) | ((val8 & 0x04U) << 3) | ((val8 & 0x02U) << 5) | ((val8 & 0x01U) << 7));
      }
    }
    switch (byteSwap) {
    case 2U:
      printf("Doing 16-bit endian-swapping...\n");
      for (i = 0; i < pBitstream->Length; i += 2) {
        val8a = pBitstream->Data[i + 0];
        val8b = pBitstream->Data[i + 1];
        pBitstream->Data[i + 0] = val8b;
        pBitstream->Data[i + 1] = val8a;
      }
      break;

    case 4U:
      printf("Doing 32-bit endian-swapping...\n");
      for (i = 0; i < pBitstream->Length; i += 2) {
        val8a = pBitstream->Data[i + 0];
        val8b = pBitstream->Data[i + 1];
        val8c = pBitstream->Data[i + 2];
        val8d = pBitstream->Data[i + 3];
        pBitstream->Data[i + 0] = val8d;
        pBitstream->Data[i + 1] = val8c;
        pBitstream->Data[i + 2] = val8b;
        pBitstream->Data[i + 3] = val8a;
      }
      break;
    }
  }

  if (bDoErase || bDoProgram) {
    /* For the 'program' or 'erase' commands, a region of the Flash is first erased  */
    printf("Erasing Flash from 0x%llX-0x%llX...\n", (unsigned long long)eraseAddress, (unsigned long long)(eraseLimit - 1U));
    status = eraseFlash(hCard, eraseAddress, eraseLimit);
    if (ADMXRC3_SUCCESS != status) {
      printf("*** Erase failed: %s", ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
  }

  if (bDoProgram) {
    /* For the 'program' commands, the bitstream is written into a region of the Flash  */
    printf("Writing bitstream image at 0x%llX-0x%llX...\n",
      (unsigned long long)programAddress, (unsigned long long)(programAddress + pBitstream->Length - 1U));
    status = programBitstream(hCard, pBitstream, programAddress);
    if (ADMXRC3_SUCCESS != status) {
      printf("*** Programming failed: %s", ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
  }

  if (bDoErase || bDoCheckBlank) {
    /* For the 'erase' commands, a blank check is performed for a region of the Flash */
    printf("Blank checking from 0x%llX-0x%llX...\n",
      (unsigned long long)eraseAddress, (unsigned long long)(eraseLimit - 1U));
    status = checkBlank(hCard, eraseAddress, eraseLimit, &nError);
    if (ADMXRC3_SUCCESS != status) {
      printf("*** Blank check failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
    if (nError) {
      printf("*** Blank check failed with %llu(0x%llX) error(s)\n",
        (unsigned long long)nError, (unsigned long long)nError);
      goto done;
    } else {
      printf("Blank check successful.\n");
    }
  }

  if (bDoProgram || bDoVerify) {
    /* For the 'program' and 'verify' commands, the bitstream data is compared against the contents of a region of the Flash */
    printf("Verifying from 0x%llX-0x%llX...\n",
      (unsigned long long)programAddress, (unsigned long long)(programAddress + pBitstream->Length - 1U));
    status = verify(hCard, pBitstream, programAddress, &nError);
    if (ADMXRC3_SUCCESS != status) {
      printf("*** Verification check failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
    if (nError) {
      printf("*** Verification failed with %llu(0x%llX) error(s)\n",
        (unsigned long long)nError, (unsigned long long)nError);
      goto done;
    } else {
      printf("Verification successful.\n");
    }
  }

  if (bDoProgram) {
    uint32_t val32;
    uint8_t val8;

    /*
    ** For the 'program' command, the flag must be set in the Flash to cause the bitstream to be loaded
    ** at reset / power up.
    */
    switch (bootFlagType) {
    case BootFlagType24BitAddress:
    case BootFlagType30BitAddress:
      /*
      ** Set the flag that causes the FPGA to be loaded from flash,
      ** and bitstream length.
      */
      if (bootFlagType == BootFlagType24BitAddress) {
        val32 = (uint32_t)((pBitstream->Length - 1U) << 8);
      } else {
        val32 = (uint32_t)((pBitstream->Length - 1U) & 0x3FFFFFFFU);
      }
      printf("Setting boot flag at 0x%llX-0x%llX to 0x%08lX...\n",
        (unsigned long long)bootFlagAddress, (unsigned long long)(bootFlagAddress + sizeof(val32) - 1U), (unsigned long)val32);
      val32 = cpu_to_le32(val32);
      status = ADMXRC3_WriteFlash(hCard, flashIndex, 0, bootFlagAddress, sizeof(val32), &val32);
      if (status != ADMXRC3_SUCCESS) {
        printf("*** Failed to set boot flag: %s\n", ADMXRC3_GetStatusString(status, TRUE));
        bBootFlagVerify = FALSE;
        ret = ERROR;
      }
      break;

    case BootFlagType8Bit:
      /* Set the flag that causes the FPGA to be loaded from flash */
      val8 = 0x0U;
      printf("Setting boot flag at 0x%llX to 0x%02X...\n", (unsigned long long)bootFlagAddress, (unsigned int)val8);
      status = ADMXRC3_WriteFlash(hCard, flashIndex, 0, bootFlagAddress, sizeof(val8), &val8);
      if (status != ADMXRC3_SUCCESS) {
        printf("*** Failed to set boot flag: %s\n", ADMXRC3_GetStatusString(status, TRUE));
        bBootFlagVerify = FALSE;
        ret = ERROR;
      }
      break;

    case BootFlagTypeNone:
      break;

    default:
      assert(FALSE);
      break;
    }
  }

  if (bDoProgram || bDoVerify) {
    if (bBootFlagVerify) {
      uint32_t expected32, bootFlag32;
      uint8_t expected8, bootFlag8;

      /*
      ** For the 'program' and 'verify' commands, the flag in the Flash that causes the bitstream to be loaded
      ** at reset / power up must be verified.
      */
      switch (bootFlagType) {
      case BootFlagType24BitAddress:
      case BootFlagType30BitAddress:
        /*
        ** Set the flag that causes the FPGA to be loaded from flash,
        ** and bitstream length.
        */
        if (bootFlagType == BootFlagType24BitAddress) {
          expected32 = (uint32_t)((pBitstream->Length - 1U) << 8);
        } else {
          expected32 = (uint32_t)((pBitstream->Length - 1U) & 0x3FFFFFFFU);
        }
        status = ADMXRC3_ReadFlash(hCard, flashIndex, 0, bootFlagAddress, sizeof(bootFlag32), &bootFlag32);
        if (status != ADMXRC3_SUCCESS) {
          printf("*** Failed to read boot flag: %s\n", ADMXRC3_GetStatusString(status, TRUE));
        } else {
          bootFlag32 = le32_to_cpu(bootFlag32);
          if (bootFlag32 != expected32) {
            printf("*** Boot flag verification failed; expecting 0x%08lX, found 0x%08lX\n",
              (unsigned long)expected32, (unsigned long)bootFlag32);
          } else {
            bBootFlagOk = TRUE;
          }
        }
        break;

      case BootFlagType8Bit:
        expected8 = 0x0U;
        status = ADMXRC3_ReadFlash(hCard, flashIndex, 0, bootFlagAddress, sizeof(bootFlag8), &bootFlag8);
        if (status != ADMXRC3_SUCCESS) {
          printf("*** Failed to read back boot flag: %s\n",
            ADMXRC3_GetStatusString(status, TRUE));
        } else {
          if (bootFlag8 != expected8) {
            printf("*** Boot flag verification failed; expecting 0x%02X, found 0x%02X\n",
              (unsigned int)expected8, (unsigned int)bootFlag8);
          } else {
            bBootFlagOk = TRUE;
          }
        }
        break;

      case BootFlagTypeNone:
        bBootFlagOk = TRUE;
        break;

      default:
        assert(FALSE);
        break;
      }
    }

    if (bootFlagType != BootFlagTypeNone) {
      if (bBootFlagOk) {
        printf("Boot flag OK. The bitstream will be loaded on reset / power up.\n");
      } else {
        printf("*** The bitstream may not be loaded on reset / power up.\n");
        ret = ERROR;
        goto done;
      }
    }
  }

  if (bDoReadback) {
    FILE* pFile;

    pFile = fopen(pBitFileName, "wb");
    if (NULL == pFile) {
      ret = ERROR;
      printf("*** Failed to open file '%s' for writing.\n", pBitFileName);
      goto done;
    }
    printf("Reading data from 0x%llX-0x%llX into '%s'...\n",
      (unsigned long long)readbackOffset, (unsigned long long)(readbackOffset + readbackLength - 1U), pBitFileName);
    status = readback(hCard, readbackOffset, readbackLength, pFile);
    if (ADMXRC3_SUCCESS != status) {
      fclose(pFile);
      ret = ERROR;
      goto done;
    }
    fclose(pFile);
  }

done:
  if (NULL != pBitstream) {
    ADMXRC3_UnloadBitstream(pBitstream);
  }

  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    if ((unsigned int)-1 != flashIndex) {
      status = ADMXRC3_SyncFlash(hCard, flashIndex, 0);
      if (ADMXRC3_SUCCESS != status) {
        printf("+++ WARNING: failed to synchronize Flash on exit: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }
    }
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return ret;
}
