/*
** monitor.c - Trivial example showing how to read sensor values from
**             the system monitoring hardware on Gen 3 Alpha Data
**             reconfigurable computing hardware.
**
** (C) Copyright 2010-2013 Alpha Data
*/

#include <vxWorks.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <admxrc3.h>
#include "sdk_common.h"
#include "platform_time.h"

#ifndef FALSE
# define FALSE (0)
#endif
#ifndef TRUE
# define TRUE (1)
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define FORMFEED (12)

static const struct {
  ADMXRC3_MODEL_TYPE model;
  const char* pName;
} models[] = {
  { ADMXRC3_MODEL_GENERIC, "Generic reconfigurable computing card" },
  { ADMXRC3_MODEL_ADMXRC2, "ADM-XRC-II" },
  { ADMXRC3_MODEL_ADPEXRC5T, "ADPe-XRC-5T" },
  { ADMXRC3_MODEL_ADMXRC6TL, "ADM-XRC-6TL" },
  { ADMXRC3_MODEL_ADMXRC6T1, "ADM-XRC-6T1" },
  { ADMXRC3_MODEL_ADMXRC6TGE, "ADM-XRC-6TGE" },
  { ADMXRC3_MODEL_ADMXRC6TADV8, "ADM-XRC-6T-ADV8" },
  { ADMXRC3_MODEL_ADPEXRC6T, "ADPE-XRC-6T" },
  { ADMXRC3_MODEL_ADPEXRC6TL, "ADPE-XRC-6T-L" },
  { ADMXRC3_MODEL_ADPEXRC6TADV_C, "ADPE-XRC-6T-ADV (Controller)" },
  { ADMXRC3_MODEL_ADPEXRC6TADV_T, "ADPE-XRC-6T-ADV (Target)" },
  { ADMXRC3_MODEL_ADMXRC7K1, "ADM-XRC-7K1" },
  { ADMXRC3_MODEL_ADMXRC6TDA1, "ADM-XRC-6T-DA1" },
  { ADMXRC3_MODEL_ADMXRC7V1, "ADM-XRC-7V1" },
  { ADMXRC3_MODEL_ADMVPX37V2, "ADM-VPX3-7V2" },
  { ADMXRC3_MODEL_ADMXRC6TGEL, "ADM-XRC-6TGEL" },
  { ADMXRC3_MODEL_ADMPCIE7V3, "ADM-PCIE-7V3" }
};

static const char*
getModelName(
  ADMXRC3_MODEL_TYPE modelType)
{
  int i;

  for (i = 0; i < ARRAY_LENGTH(models); i++) {
    if (models[i].model == modelType) {
      return models[i].pName;
    }
  }

  return NULL;
}

static const char*
getUnit(
  ADMXRC3_UNIT_TYPE unit)
{
  switch (unit) {
  case ADMXRC3_UNIT_NONE:
    return "";

  case ADMXRC3_UNIT_A:
    return "A";

  case ADMXRC3_UNIT_V:
    return "V";

  case ADMXRC3_UNIT_C:
    return "deg C";

  case ADMXRC3_UNIT_HZ:
    return "Hz";

  case ADMXRC3_UNIT_RPM:
    return "RPM";

  case ADMXRC3_UNIT_S:
    return "s";

  default:
    return "(unknown unit)";
  }
}

#define FLAG_BYSERIAL (0x1U)  /* Interpret 'index' as a serial number */

STATUS
admxrc3Monitor(
  unsigned int index,  /* Index of device to open (OR serial number if 'flags' includes BYSERIAL) */
  unsigned int flags,  /* See above flags */
  unsigned int period, /* Update period, in seconds */
  unsigned int numRep) /* Number of updates to perform (0 => infinite) */
{
  STATUS ret = OK;
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_CARD_INFOEX cardInfoEx;
  ADMXRC3_SENSOR_INFO sensorInfo;
  char modelName[80];
  unsigned int i, rep;
  bool_t bOpenBySerial;
  
  sampleTaskNice();

  bOpenBySerial = (flags & FLAG_BYSERIAL) ? TRUE : FALSE;

  status = sampleOpenCard(index, !bOpenBySerial, index, bOpenBySerial, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index/serial %lu: %s\n",
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  if (period < 1U) {
    period = 1U;
  }
  if (period > 100U) {
    period = 100U;
  }

  status = ADMXRC3_GetCardInfoEx(hCard, &cardInfoEx);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }
  if (NULL != getModelName(cardInfoEx.Model)) {
    strcpy(modelName, getModelName(cardInfoEx.Model));
  } else {
    sprintf(modelName, "unknown (0x%lX)", (unsigned long)cardInfoEx.Model);
  }

  for (rep = 0; numRep == 0 || rep < numRep; rep++) {
    ADMXRC3_SENSOR_VALUE sensorValue;

    if (rep >= 1) {
      sleepMs(period * 1000);
    }

    printf("Model:                                    %lu (0x%lX) => %s \n", (unsigned long)cardInfoEx.Model, (unsigned long)cardInfoEx.Model, modelName);
    printf("Serial number:                            %lu (0x%lX)\n", (unsigned long)cardInfoEx.SerialNumber, (unsigned long)cardInfoEx.SerialNumber);
    printf("Number of sensors:                        %lu\n", (unsigned long)cardInfoEx.NumSensor);

    if (0 == cardInfoEx.NumSensor) {
      printf("*** No sensors to monitor. Exiting.\n");
      ret = ERROR;
      goto done;
    }

    for (i = 0; i < cardInfoEx.NumSensor; i++) {
      printf("  Sensor %2u ", i);
      status = ADMXRC3_GetSensorInfo(hCard, i, &sensorInfo);
      if (status != ADMXRC3_SUCCESS) {
        printf("(Failed to get info: %s)\n", ADMXRC3_GetStatusString(status, TRUE));
        continue;
      }

      status = ADMXRC3_ReadSensor(hCard, i, &sensorValue);
      if (status != ADMXRC3_SUCCESS) {
        printf("%28s: (Failed to read sensor: %s)\n", sensorInfo.Description, ADMXRC3_GetStatusString(status, TRUE));
        continue;
      }

      switch (sensorInfo.DataType) {
      case ADMXRC3_DATA_BOOL:
        assert(sensorInfo.Exponent == 0 && sensorInfo.Unit == ADMXRC3_UNIT_NONE);
        printf("%28s: %s\n", sensorInfo.Description, sensorValue.Bool ? "TRUE" : "FALSE");
        break;

      case ADMXRC3_DATA_DOUBLE:
        printf("%28s: %f", sensorInfo.Description, sensorValue.Double);
        break;

      case ADMXRC3_DATA_INT32:
        printf("%28s: %ld", sensorInfo.Description, (long)sensorValue.Int32);
        break;

      case ADMXRC3_DATA_UINT32:
        printf("%28s: %lu", sensorInfo.Description, (unsigned long)sensorValue.UInt32);
        break;

      default:
        /* Unknown type; can't interpret */
        printf("%28s: ?", sensorInfo.Description);
        break;
      }
      if (sensorInfo.DataType != ADMXRC3_DATA_BOOL) {
        if (sensorInfo.Exponent) {
          printf(" * 10e%d %s\n", sensorInfo.Exponent, getUnit(sensorInfo.Unit));
        } else {
          printf(" %s\n", getUnit(sensorInfo.Unit));
        }
      }
    }

    printf("\n");
  }

done:
  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return ret;
}
