/*
** vpd.c - utility to display / modify Vital Product Data for an ADM-XRC
**         series card.
**
** (C) Copyright 2009-2013 Alpha Data
**
** NOTE: for fill and write commands to succeed, VPD write-protection must be
**       disabled as detailed in the ADB3 Driver for VxWorks release notes.
*/

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>

#include "sdk_common.h"
#include "platform_endian.h"

#ifndef FALSE
# define FALSE (0)
#endif
#ifndef TRUE
# define TRUE (1)
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

typedef enum _CommandType {
  CommandFill,
  CommandRead,
  CommandWrite
} CommandType;

typedef enum _TokenStateType {
  TokenStateInToken,
  TokenStateInSpace
} TokenStateType;

typedef enum _EscapeStateType {
  EscapeStateNormal,
  EscapeStateInEscape
} EscapeStateType;

typedef struct _State {
  TokenStateType state;
  EscapeStateType escapeState;
  unsigned int tokenPos;
  char tokenBuf[128];
} State;

static const char g_funcName[] = "admxrc3Vpd";

static const State g_initialState = {
  TokenStateInSpace,
  EscapeStateNormal,
  0
};

static int
ourStrCaseCmp(
  const char* pS1,
  const char* pS2)
{
  char c1, c2;

  while (1) {
    c1 = *pS1;
    c2 = *pS2;
    if (c1 == '\0' || tolower(c1) != tolower(c2)) {
      break;
    }
    pS1++;
    pS2++;
  }
  return tolower(c1) - tolower(c2);
}

static void
strset_s(
  char* pBuffer,
  size_t max,
  char c,
  size_t n)
{
  if (!max) {
    return;
  }
  max--;
  while (max) {
    *pBuffer++ = c;
    max--;
  }
  *pBuffer++ = '\0';
}

static size_t
remaining(
  const char* pPos,
  const char* line,
  size_t max)
{
  size_t n = pPos - line;
  if (n > max) {
    return 0;
  } else {
     return max - n;
  }
}

static void
showUsage(
  const char* pFuncName)
{
  printf("Usage:\n");
  printf("    %s <index>, <flags>, \"<fill-command>\", <address>, <n>, \"<data> ...\"\n", pFuncName);
  printf("    %s <index>, <flags>, \"<read-command>\", <address>, <n>\n", pFuncName);
  printf("    %s <index>, <flags>, \"<write-command>\", <address>, <n>, [\"<data> ...\"]\n", pFuncName);
  printf("where <flags> is one or more of:\n");
  printf("    0x1 => BYSERIAL  interpret <index> as serial number of card to open\n");
  printf("    0x2 => HEX       interpret <data> as hexadecimal (fill & write commands)\n");
  printf("where <fill-command> is one of:\n");
  printf("    fb, fw, fd, fq, fs => fill BYTEs, WORDs, DWORDs, QWORDs, string\n");
  printf("where <read-command> is one of:\n");
  printf("    rb, rw, rd, rq => read BYTEs, WORDs, DWORDs, QWORDs\n");
  printf("where <write-command> is one of:\n");
  printf("    wb, ww, wd, wq, ws => write BYTEs, WORDs, DWORDs, QWORDs, string\n");
}

static ADMXRC3_STATUS
dumpVpd(
  ADMXRC3_HANDLE hCard,
  unsigned int indent,
  unsigned int nColumn,
  unsigned int radix,
  bool_t bAlign,
  uint32_t offset,
  size_t n)
{
  ADMXRC3_STATUS status;
  uint32_t address, limit, addressCol, m;
  unsigned int addressLo, columnMask;
  uint32_t tmpOffs;
  uint8_t* p8;
  char line[257];
  char* pPos;
  uint8_t columnBuf[256];
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;
  unsigned int i, j, k, chunk;
  unsigned int rowBytes;
  
  if (n > (size_t)0xffffffffU) {
    printf("*** dumpVpd: invalid value '0x%lx' for n\n", (unsigned long)n);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  if (indent > 128) {
    printf("*** dumpVpd: invalid value '%lu' for indent\n", (unsigned long)indent);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  switch (nColumn) {
  case 1:
  case 2:
  case 4:
  case 8:
  case 16:
  case 32:
    break;
  default:
    printf("*** dumpVpd: invalid value '%lu' for nColumn\n", (unsigned long)nColumn);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  if (radix != 1 && radix != 2 && radix != 4 && radix != 8) {
    printf("*** dumpVpd: invalid value '%lu' for radix\n", (unsigned long)radix);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  
  rowBytes = nColumn * radix;

  /* Make sure everything is aligned */
  columnMask = rowBytes - 1U;
  address = offset;
  limit = address + (uint32_t)n;
  if (bAlign) {
    addressCol = address & ~(uint32_t)columnMask;
    addressLo = ((unsigned int)address) & columnMask;
  } else {
    addressCol = address;
    addressLo = 0;
  }
  
  m = limit - address;
  tmpOffs = offset;
  
  pPos = line;
  strset_s(pPos, remaining(pPos, line, sizeof(line) / sizeof(line[0])), ' ', indent);
  pPos += indent;
#if (_WRS_VXWORKS_MAJOR >= 6)
  snprintf(pPos, remaining(pPos, line, sizeof(line) / sizeof(line[0])), "Dump of VPD at 0x%lx + %lu(0x%lx) bytes:\n",
    (unsigned long)offset, (unsigned long)n, (unsigned long)n);
#else
  /* Need VxWorks 6.x for snprintf() */
  sprintf(pPos, "Dump of VPD at 0x%lx + %lu(0x%lx) bytes:\n", (unsigned long)offset, (unsigned long)n, (unsigned long)n);
#endif
  printf("%s", line);
  
  pPos = line;
  strset_s(pPos, remaining(pPos, line, sizeof(line) / sizeof(line[0])), ' ', indent + 12);
  pPos += indent + 12;
  for (i = 0; i < rowBytes; i += radix) {
    switch (radix) {
    case 0x1U:
#if (_WRS_VXWORKS_MAJOR >= 6)
      snprintf(pPos, remaining(pPos, line, sizeof(line)), "%02lx ", (unsigned long)i);
#else
      /* Need VxWorks 6.x for snprintf() */
      sprintf(pPos, "%02lx ", (unsigned long)i);
#endif
      pPos += 3;
      break;
      
    case 0x2U:
#if (_WRS_VXWORKS_MAJOR >= 6)
      snprintf(pPos, remaining(pPos, line, sizeof(line)), "  %02lx ", (unsigned long)i);
#else
      /* Need VxWorks 6.x for snprintf() */
      sprintf(pPos, "  %02lx ", (unsigned long)i);
#endif
      pPos += 5;
      break;
      
    case 0x4U:
#if (_WRS_VXWORKS_MAJOR >= 6)
      snprintf(pPos, remaining(pPos, line, sizeof(line)), "      %02lx ", (unsigned long)i);
#else
      /* Need VxWorks 6.x for snprintf() */
      sprintf(pPos, "      %02lx ", (unsigned long)i);
#endif
      pPos += 9;
      break;
      
    case 0x8U:
#if (_WRS_VXWORKS_MAJOR >= 6)
      snprintf(pPos, remaining(pPos, line, sizeof(line)), "               %02lx ", (unsigned long)i);
#else
      /* Need VxWorks 6.x for snprintf() */
      sprintf(pPos, "               %02lx ", (unsigned long)i);
#endif
      pPos += 18;
      break;
    }
  }
  *pPos++ = '\n';
  *pPos++ = '\0';
  printf("%s", line);
  
  while (m) {
    chunk = rowBytes - addressLo;
    if (chunk > m) {
      chunk = (unsigned int)m;
    }
    
    memset(columnBuf, 0, rowBytes);
    status = ADMXRC3_ReadVPD(hCard, 0, tmpOffs, chunk, columnBuf + addressLo);
    if (ADMXRC3_SUCCESS != status) {
      printf("*** Failed to read VPD chunk at offset 0x%lx, size 0x%lx: %s\n",
        (unsigned long)tmpOffs, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    p8 = (uint8_t*)columnBuf;
    
    pPos = line;
    strset_s(pPos, remaining(pPos, line, sizeof(line) / sizeof(line[0])), ' ', indent);
    pPos += indent;
#if (_WRS_VXWORKS_MAJOR >= 6)
    snprintf(pPos, remaining(pPos, line, sizeof(line)), "0x%08lx: ", (unsigned long)(uint32_t)addressCol);
#else
    /* Need VxWorks 6.x for snprintf() */
    sprintf(pPos, "0x%08lx: ", (unsigned long)(uint32_t)addressCol);
#endif
    pPos += 12;
    
    j = addressLo + chunk;
    for (i = 0; i < rowBytes; i += radix) {
      switch (radix) {
      case 0x1U:
        val8 = *p8++;
        if (i >= addressLo && i < j) {
#if (_WRS_VXWORKS_MAJOR >= 6)
          snprintf(pPos, remaining(pPos, line, sizeof(line)), "%02lx", (unsigned long)val8);
#else
          /* Need VxWorks 6.x for snprintf() */
          sprintf(pPos, "%02lx", (unsigned long)val8);
#endif
          pPos += 2;
        } else {
          *pPos++ = '.';
          *pPos++ = '.';
        }
        break;
        
      case 0x2U:
        val16 = (uint16_t)(((uint16_t)p8[0] << 0U) | ((uint16_t)p8[1] << 8U));
        p8 += 2;
        for (k = 2; k != 0;) {
          k--;
          if (i + k >= addressLo && i + k < j) {
#if (_WRS_VXWORKS_MAJOR >= 6)
            snprintf(pPos, remaining(pPos, line, sizeof(line)), "%02lx", (unsigned long)(uint8_t)(val16 >> (k << 3)));
#else
            /* Need VxWorks 6.x for snprintf() */
            sprintf(pPos, "%02lx", (unsigned long)(uint8_t)(val16 >> (k << 3)));
#endif
            pPos += 2;
          } else {
            *pPos++ = '.';
            *pPos++ = '.';
          }
          if (k == 3) {
            *pPos++ = '_';
          }
        }
        break;
        
      case 0x4U:
        val32 = (((uint32_t)p8[0]) <<  0U) | (((uint32_t)p8[1]) <<  8U) | (((uint32_t)p8[2]) << 16U) | (((uint32_t)p8[3]) << 24U);
        p8 += 4;
        for (k = 4; k != 0;) {
          k--;
          if (i + k >= addressLo && i + k < j) {
#if (_WRS_VXWORKS_MAJOR >= 6)
            snprintf(pPos, remaining(pPos, line, sizeof(line)), "%02lx", (unsigned long)(uint8_t)(val32 >> (k << 3)));
#else
            /* Need VxWorks 6.x for snprintf() */
            sprintf(pPos, "%02lx", (unsigned long)(uint8_t)(val32 >> (k << 3)));
#endif
            pPos += 2;
          } else {
            *pPos++ = '.';
            *pPos++ = '.';
          }
        }
        break;
        
      case 0x8U:
        val64  = (((uint64_t)p8[0]) <<  0U) | (((uint64_t)p8[1]) <<  8U) | (((uint64_t)p8[2]) << 16U) | (((uint64_t)p8[3]) << 24U);
        val64 |= (((uint64_t)p8[4]) << 32U) | (((uint64_t)p8[5]) << 40U) | (((uint64_t)p8[6]) << 48U) | (((uint64_t)p8[7]) << 56U);
        p8 += 8;
        for (k = 8; k != 0;) {
          k--;
          if (i + k >= addressLo && i + k < j) {
#if (_WRS_VXWORKS_MAJOR >= 6)
            snprintf(pPos, remaining(pPos, line, sizeof(line)), "%02lx", (unsigned long)(uint8_t)(val64 >> (k << 3)));
#else
            /* Need VxWorks 6.x for snprintf() */
            sprintf(pPos, "%02lx", (unsigned long)(uint8_t)(val64 >> (k << 3)));
#endif
            pPos += 2;
          } else {
            *pPos++ = '.';
            *pPos++ = '.';
          }
          if (k == 4) {
            *pPos++ = '_';
          }
        }
        break;
      }
      *pPos++ = ' ';
    }
    
    p8 = (uint8_t*)columnBuf;
    for (i = 0; i < rowBytes; i++) {
      char c = (char)*p8++;
      
      if (isprint((int)c)) {
        *pPos++ = c;
      } else {
        *pPos++ = '.';
      }
    }
    *pPos++ = '\n';
    *pPos++ = '\0';
    printf("%s", line);
    addressLo = 0;
    addressCol += rowBytes;
    m -= chunk;
    tmpOffs += chunk;
  }
  
  return ADMXRC3_SUCCESS;
}

static ADMXRC3_STATUS
fillVpd(
  ADMXRC3_HANDLE hCard,
  uint32_t offset,
  unsigned int count,
  unsigned int width,
  bool_t bInterpretString,
  bool_t bInterpretHex,
  const char* pString)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint8_t buf[65536];
  const unsigned int bufSize = ARRAY_LENGTH(buf);
  unsigned int bufPos, strPos;
  size_t length;
  unsigned int chunk;
  unsigned long long tmp;
  uint8_t* p8;
  uint16_t* p16;
  uint32_t* p32;
  uint64_t* p64;
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;

  length = strlen(pString);
  if (bInterpretString) {
    strPos = 0;
    while (count) {
      chunk = (count > bufSize) ? bufSize : count;
      bufPos = 0;
      while (bufPos < bufSize && bufPos < chunk) {
        buf[bufPos] = pString[strPos];
        bufPos++;
        strPos++;
        if (strPos >= length) {
          strPos = 0;
        }
      }
      status = ADMXRC3_WriteVPD(hCard, 0, offset, chunk, buf);
      if (ADMXRC3_SUCCESS != status) {
        printf("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n",
          (unsigned long)offset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
        return status;
      }
      offset += chunk;
      count -= chunk;
    }
  } else {
    if (length >= 2 && pString[0] == '0' && (pString[1] == 'x' || pString[1] == 'X')) {
      sscanf(pString + 2, "%llx", &tmp);
    } else {
      if (bInterpretHex) {
        sscanf(pString, "%llx", &tmp);
      } else {
        sscanf(pString, "%llu", &tmp);
      }
    }
    val64 = (uint64_t)tmp;
    switch (width) {
      case 1U:
        val8 = (uint8_t)val64;
        p8 = (uint8_t*)buf;
        while (count) {
          chunk = (count > bufSize) ? bufSize : count;
          bufPos = 0;
          while (bufPos < chunk) {
            *p8++ = val8;
            bufPos++;
          }
          status = ADMXRC3_WriteVPD(hCard, 0, offset, chunk, buf);
          if (ADMXRC3_SUCCESS != status) {
            printf("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n",
              (unsigned long)offset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
            return status;
          }
          offset += chunk;
          count -= chunk;
        }
        break;

      case 2U:
        val16 = (uint16_t)val64;
        p16 = (uint16_t*)buf;
        while (count) {
          chunk = (count > bufSize) ? bufSize : count;
          bufPos = 0;
          while (bufPos < chunk) {
            *p16++ = cpu_to_le16(val16);
            bufPos += 2;
          }
          status = ADMXRC3_WriteVPD(hCard, 0, offset, chunk, buf);
          if (ADMXRC3_SUCCESS != status) {
            printf("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n",
              (unsigned long)offset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
            return status;
          }
          offset += chunk;
          count -= chunk;
        }
        break;

      case 4U:
        val32 = (uint32_t)val64;
        p32 = (uint32_t*)buf;
        while (count) {
          chunk = (count > bufSize) ? bufSize : count;
          bufPos = 0;
          while (bufPos < chunk) {
            *p32++ = cpu_to_le32(val32);
            bufPos += 4;
          }
          status = ADMXRC3_WriteVPD(hCard, 0, offset, chunk, buf);
          if (ADMXRC3_SUCCESS != status) {
            printf("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n",
              (unsigned long)offset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
            return status;
          }
          offset += chunk;
          count -= chunk;
        }
        break;

      case 8U:
        p64 = (uint64_t*)buf;
        while (count) {
          chunk = (count > bufSize) ? bufSize : count;
          bufPos = 0;
          while (bufPos < chunk) {
            *p64++ = cpu_to_le64(val64);
            bufPos += 8;
          }
          status = ADMXRC3_WriteVPD(hCard, 0, offset, chunk, buf);
          if (ADMXRC3_SUCCESS != status) {
            printf("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n",
              (unsigned long)offset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
            return status;
          }
          offset += chunk;
          count -= chunk;
        }
        break;

      default:
        assert(FALSE);
        break;
    }
  }

  return status;
}

static ADMXRC3_STATUS
writeVpdToken(
  ADMXRC3_HANDLE hCard,
  uint32_t* pOffset,
  unsigned int* pCount,
  unsigned int width,
  bool_t bInterpretString,
  bool_t bInterpretHex,
  const char* pString,
  size_t length)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint8_t buf[65536];
  const unsigned int bufSize = ARRAY_LENGTH(buf);
  uint64_t val64;
  uint32_t val32;
  uint16_t val16;
  uint8_t val8;
  unsigned long long tmp;
  unsigned int i, chunk;

  if (bInterpretString) {
    while (*pCount) {
      if (length > *pCount) {
        chunk = *pCount;
      } else {
        chunk = (unsigned int)length; /* Cast is safe due to above check */
      }
      if (chunk > bufSize) {
        chunk = bufSize;
      }
      for (i = 0; i < chunk; i++) {
        buf[i] = *pString++;
      }
      status = ADMXRC3_WriteVPD(hCard, 0, *pOffset, chunk, buf);
      if (ADMXRC3_SUCCESS != status) {
        printf("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n",
          (unsigned long)*pOffset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
        return status;
      }
      *pOffset += chunk;
      *pCount -= chunk;
    }
  } else {
    if (strlen(pString) >= 2 && pString[0] == '0' && (pString[1] == 'x' || pString[1] == 'X')) {
      sscanf(pString + 2, "%llx", &tmp);
    } else {
      if (bInterpretHex) {
        sscanf(pString, "%llx", &tmp);
      } else {
        sscanf(pString, "%llu", &tmp);
      }
    }
    val64 = (uint64_t)tmp;
    switch (width) {
    case 1U:
      val8 = (uint8_t)val64;
      status = ADMXRC3_WriteVPD(hCard, 0, *pOffset, width, &val8);
      break;

    case 2U:
      val16 = cpu_to_le16((uint16_t)val64);
      status = ADMXRC3_WriteVPD(hCard, 0, *pOffset, width, &val16);
      break;

    case 4U:
      val32 = cpu_to_le32((uint32_t)val64);
      status = ADMXRC3_WriteVPD(hCard, 0, *pOffset, width, &val32);
      break;

    case 8U:
      val64 = cpu_to_le64(val64);
      status = ADMXRC3_WriteVPD(hCard, 0, *pOffset, width, &val64);
      break;

    default:
      assert(FALSE);
      break;
    }
    if (ADMXRC3_SUCCESS != status) {
      printf("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n",
        (unsigned long)*pOffset, (unsigned long)width, ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }
    *pOffset += width;
    *pCount -= width;
  }

  return status;
}

static ADMXRC3_STATUS
writeVpd(
  ADMXRC3_HANDLE hCard,
  uint32_t *pOffset,
  unsigned int* pCount,
  State* pState,
  unsigned int width,
  bool_t bInterpretString,
  bool_t bInterpretHex,
  const char* pString)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  char c;

  while (*pCount) {
    c = *pString++;
    if (c == '\0') {
      break;
    }
    if (bInterpretString) {
      switch (pState->state) {
      case TokenStateInToken:
        if (c == '\n') {
          status = writeVpdToken(hCard, pOffset, pCount, width, bInterpretString, bInterpretHex, pState->tokenBuf, pState->tokenPos);
          if (ADMXRC3_SUCCESS != status) {
            return status;
          }
          pState->state = TokenStateInSpace;
          pState->tokenPos = 0;
        } else {
          pState->tokenBuf[pState->tokenPos] = c;
          pState->tokenPos++;
        }
        break;

      case TokenStateInSpace:
        if (c != '\n') {
          pState->state = TokenStateInToken;
          pState->tokenBuf[pState->tokenPos] = c;
          pState->tokenPos++;
        }
        break;

      default:
        assert(FALSE);
        break;
      }
    } else {
      switch (pState->state) {
      case TokenStateInToken:
        if (isspace((int)c)) {
          status = writeVpdToken(hCard, pOffset, pCount, width, bInterpretString, bInterpretHex, pState->tokenBuf, pState->tokenPos);
          if (ADMXRC3_SUCCESS != status) {
            return status;
          }
          pState->state = TokenStateInSpace;
          pState->tokenPos = 0;
        } else {
          pState->tokenBuf[pState->tokenPos] = c;
          pState->tokenPos++;
        }
        break;

      case TokenStateInSpace:
        if (!isspace((int)c)) {
          pState->state = TokenStateInToken;
          pState->tokenBuf[pState->tokenPos] = c;
          pState->tokenPos++;
        }
        break;

      default:
        assert(FALSE);
        break;
      }
    }
  }

  return status;
}

/* Bits for 'flags' parameter */
#define FLAG_BYSERIAL (0x1U)   /* Interpret 'index' as a serial number */
#define FLAG_HEX      (0x10U)  /* For fill and write commands, interpret values as hexadecimal */

STATUS
admxrc3Vpd(
  unsigned int index,          /* Index of device to open (OR serial number if 'flags' includes BYSERIAL) */
  unsigned int flags,          /* See above flags */
  const char* pCommand,        /* "fb", "fw", "rd", "fq", "fs", "rb", "rw", "rd", "rq", "wb", "ww", "wd", "wq" or "ws" */
  unsigned int startOffset,
  unsigned int n,
  const char* pArg)            /* Data items, valid for fill and write commands */
{
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  STATUS ret = OK;
  CommandType command;
  bool_t bOpenBySerial, bInterpretString, bInterpretHex;
  unsigned int width;
  uint32_t offset = (uint32_t)startOffset;
  int c;
  char str[128];
  State state = g_initialState;

  if (pCommand == NULL) {
    printf("*** Please specify a valid command.\n");
    showUsage(g_funcName);
    ret = ERROR;
    goto done;
  }

  bOpenBySerial = (flags & FLAG_BYSERIAL) ? TRUE : FALSE;
  bInterpretHex = (flags & FLAG_HEX) ? TRUE : FALSE;
  bInterpretString = FALSE;
  if (ourStrCaseCmp(pCommand, "rb") == 0) {
    width = 1;
    command = CommandRead;
  } else if (ourStrCaseCmp(pCommand, "rw") == 0) {
    width = 2;
    command = CommandRead;
  } else if (ourStrCaseCmp(pCommand, "rd") == 0) {
    width = 4;
    command = CommandRead;
  } else if (ourStrCaseCmp(pCommand, "rq") == 0) {
    width = 8;
    command = CommandRead;
  } else if (ourStrCaseCmp(pCommand, "wb") == 0) {
    width = 1;
    command = CommandWrite;
  } else if (ourStrCaseCmp(pCommand, "ww") == 0) {
    width = 2;
    command = CommandWrite;
  } else if (ourStrCaseCmp(pCommand, "wd") == 0) {
    width = 4;
    command = CommandWrite;
  } else if (ourStrCaseCmp(pCommand, "wq") == 0) {
    width = 8;
    command = CommandWrite;
  } else if (ourStrCaseCmp(pCommand, "ws") == 0) {
    width = 1;
    bInterpretString = TRUE;
    command = CommandWrite;
  } else if (ourStrCaseCmp(pCommand, "fb") == 0) {
    width = 1;
    command = CommandFill;
  } else if (ourStrCaseCmp(pCommand, "fw") == 0) {
    width = 2;
    command = CommandFill;
  } else if (ourStrCaseCmp(pCommand, "fd") == 0) {
    width = 4;
    command = CommandFill;
  } else if (ourStrCaseCmp(pCommand, "fq") == 0) {
    width = 8;
    command = CommandFill;
  } else if (ourStrCaseCmp(pCommand, "fs") == 0) {
    width = 1;
    bInterpretString = TRUE;
    command = CommandFill;
  } else {
    printf("*** Invalid command: %s\n", pCommand);
    showUsage(g_funcName);
    ret = ERROR;
    goto done;
  }

  if (0 == n) {
    n = width;
  }
  if (n % width) {
    printf("+++ Warning: byte count %u(0x%x) is not an integer multiple of width (%u)\n", n, n, width);
    n &= ~(width - 1U);
    printf("+++          byte count adjusted to %u(0x%x)\n", n, n);
  }

  status = sampleOpenCard(index, !bOpenBySerial, index, bOpenBySerial, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index/serial %lu: %s\n",
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  switch (command) {
  case CommandFill:
    if (NULL == pArg) {
      printf("*** Too few arguments for fill command.\n");
      ret = ERROR;
      goto done;
    }
    status = fillVpd(hCard, offset, n, width, bInterpretString, bInterpretHex, pArg);
    if (ADMXRC3_SUCCESS != status) {
      ret = ERROR;
      goto done;
    }
    break;

  case CommandRead:
    status = dumpVpd(hCard, 0, 16 / width, width, TRUE, offset, n);
    if (ADMXRC3_SUCCESS != status) {
      ret = ERROR;
      goto done;
    }
    break;

  case CommandWrite:
    if (NULL != pArg) {
      printf("0x%08lx: %s\n", (unsigned long)offset, pArg);
      status = writeVpd(hCard, &offset, &n, &state, width, bInterpretString, bInterpretHex, pArg);
      if (status != ADMXRC3_SUCCESS) {
        ret = ERROR;
        goto done;
      }
      /* Complete current token (if necessary) */ 
      status = writeVpd(hCard, &offset, &n, &state, width, bInterpretString, bInterpretHex, "\n");
      if (status != ADMXRC3_SUCCESS) {
        ret = ERROR;
        goto done;
      }
    }
    if (0 != n) {
      printf("0x%08lx: ", (unsigned long)offset);
      fflush(stdout);
    }
    while (n) {
      c = getc(stdin);
      if (c == EOF) {
        /* Complete last token (if necessary) */
        status = writeVpd(hCard, &offset, &n, &state, width, bInterpretString, bInterpretHex, "\n");
        if (status != ADMXRC3_SUCCESS) {
          ret = ERROR;
          goto done;
        }
        break;
      }
      str[0] = (char)c;
      str[1] = '\0';
      status = writeVpd(hCard, &offset, &n, &state, width, bInterpretString, bInterpretHex, str);
      if (status != ADMXRC3_SUCCESS) {
        ret = ERROR;
        goto done;
      }
      if (c == '\n' && 0 != n) {
        printf("0x%08lx: ", (unsigned long)offset);
        fflush(stdout);
      }
    }
    break;

  default:
    assert(FALSE);
    break;
  }

done:
  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return ret;
}
